# -*- coding: utf-8 -*-

import pytest
import numpy as np

from gordo_components.server import utils as server_utils
import tests.utils as tu


@pytest.mark.parametrize(
    "data_to_post",
    [
        {
            "X": np.random.random(size=(10, len(tu.SENSORS_STR_LIST))).tolist(),
            "y": np.random.random(size=(10, len(tu.SENSORS_STR_LIST))).tolist(),
        },  # Nested records
        {
            "X": np.random.random(size=(1, len(tu.SENSORS_STR_LIST))).tolist(),
            "y": np.random.random(size=(1, len(tu.SENSORS_STR_LIST))).tolist(),
        },  # Single record
    ],
)
@pytest.mark.parametrize("resp_format", ("json", "parquet", None))
def test_anomaly_prediction_endpoint(
    base_route, influxdb, gordo_ml_server_client, data_to_post, sensors, resp_format
):
    """
    Anomaly GET and POST responses are the same
    """
    endpoint = f"{base_route}/anomaly/prediction"
    if resp_format is not None:
        endpoint += f"?format={resp_format}"

    resp = gordo_ml_server_client.post(endpoint, json=data_to_post)

    # From here, the response should be (pretty much) the same format from GET or POST
    assert resp.status_code == 200
    if resp_format in (None, "json"):
        assert "data" in resp.json
        data = server_utils.dataframe_from_dict(resp.json["data"])
    else:
        data = server_utils.dataframe_from_parquet_bytes(resp.data)

    # Only different between POST and GET is POST will return None for
    # start and end dates, because the server can't know what those are
    assert "start" in data
    assert "end" in data
    if data_to_post is not None:
        assert np.all(data["start"].isna())
        assert np.all(data["end"].isna())
    else:
        assert not np.any(data["start"].isna())
        assert not np.any(data["end"].isna())

    assert all(
        key in data
        for key in ("total-anomaly", "tag-anomaly", "model-input", "model-output")
    )



# -*- coding: utf-8 -*-

import os
import json
import logging
import tempfile
import typing
from dateutil.parser import isoparse  # type: ignore

import pytest
import requests
import pandas as pd
import numpy as np
from unittest import mock
from click.testing import CliRunner
from sklearn.base import BaseEstimator
from mock import patch, call

from gordo_components.client import Client, utils as client_utils
from gordo_components.client.utils import EndpointMetadata
from gordo_components.client.io import (
    _handle_response,
    HttpUnprocessableEntity,
    BadRequest,
)
from gordo_components.client.forwarders import ForwardPredictionsIntoInflux
from gordo_components.client.utils import PredictionResult
from gordo_components.data_provider import providers
from gordo_components.server import utils as server_utils
from gordo_components.model import utils as model_utils
from gordo_components import cli, serializer
from gordo_components.cli import custom_types

from tests import utils as tu


def test_client_get_metadata(watchman_service):
    """
    Test client's ability to get metadata from some target
    """
    client = Client(project=tu.GORDO_PROJECT)

    metadata = client.get_metadata()
    assert isinstance(metadata, dict)

    # Can't get metadata for non-existent target
    with pytest.raises(ValueError):
        client = Client(project=tu.GORDO_PROJECT, target="no-such-target")
        client.get_metadata()


def test_client_predict_specific_targets(watchman_service):
    """
    Client.predict should filter any endpoints given to it.
    """
    client = Client(project=tu.GORDO_PROJECT)
    with mock.patch.object(
        Client,
        "predict_single_endpoint",
        return_value=PredictionResult("test-name", [], []),
    ) as patched:

        start = (isoparse("2016-01-01T00:00:00+00:00"),)
        end = isoparse("2016-01-01T12:00:00+00:00")

        # Should not actually call any predictions because this machine name doesn't exist
        client.predict(start=start, end=end, endpoint_names=["non-existant-machine"])
        patched.assert_not_called()

        # Should be called once, for this machine.
        client.predict(start=start, end=end, endpoint_names=[tu.GORDO_SINGLE_TARGET])
        patched.assert_called_once()


def test_client_download_model(watchman_service):
    """
    Test client's ability to download the model
    """
    client = Client(project=tu.GORDO_PROJECT, target=tu.GORDO_SINGLE_TARGET)

    models = client.download_model()
    assert isinstance(models, dict)
    assert isinstance(models[tu.GORDO_SINGLE_TARGET], BaseEstimator)

    # Can't download model for non-existent target
    with pytest.raises(ValueError):
        client = Client(project=tu.GORDO_PROJECT, target="non-existent-target")
        client.download_model()


@pytest.mark.parametrize("batch_size", (10, 100))
@pytest.mark.parametrize("use_parquet", (True, False))
def test_client_predictions_diff_batch_sizes(
    influxdb, watchman_service, batch_size: int, use_parquet: bool
):
    """
    Run the prediction client with different batch-sizes and whether to use
    a data provider or not.
    """
    # Time range used in this test
    start, end = (
        isoparse("2016-01-01T00:00:00+00:00"),
        isoparse("2016-01-01T12:00:00+00:00"),
    )

    # Client only used within the this test
    test_client = client_utils.influx_client_from_uri(tu.INFLUXDB_URI)

    # Created measurements by prediction client with dest influx
    query = f"""
    SELECT *
    FROM "model-output"
    WHERE("machine" =~ /^{tu.GORDO_SINGLE_TARGET}$/)
    """

    # Before predicting, influx destination db should be empty for 'predictions' measurement
    vals = test_client.query(query)
    assert len(vals) == 0

    data_provider = providers.InfluxDataProvider(
        measurement=tu.INFLUXDB_MEASUREMENT,
        value_name="Value",
        client=client_utils.influx_client_from_uri(
            uri=tu.INFLUXDB_URI, dataframe_client=True
        ),
    )

    prediction_client = Client(
        project=tu.GORDO_PROJECT,
        data_provider=data_provider,
        prediction_forwarder=ForwardPredictionsIntoInflux(  # type: ignore
            destination_influx_uri=tu.INFLUXDB_URI
        ),
        batch_size=batch_size,
        use_parquet=use_parquet,
        parallelism=10,
    )

    # Should have discovered machine-1
    assert len(prediction_client.endpoints) == 1

    # All endpoints should be healthy
    assert all(ep.healthy for ep in prediction_client.endpoints)

    # Get predictions
    predictions = prediction_client.predict(start=start, end=end)
    assert isinstance(predictions, list)
    assert len(predictions) == 1

    name, predictions, error_messages = predictions[0]  # First dict of predictions
    assert isinstance(name, str)
    assert isinstance(predictions, pd.DataFrame)
    assert isinstance(error_messages, list)

    assert isinstance(predictions.index, pd.core.indexes.datetimes.DatetimeIndex)

    # This should have resulted in writting predictions to influx
    # Before predicting, influx destination db should be empty
    vals = test_client.query(query)
    assert (
        len(vals) > 0
    ), f"Expected new values in 'predictions' measurement, but found {vals}"


@pytest.mark.parametrize(
    "args",
    [
        ["client", "--help"],
        ["client", "predict", "--help"],
        ["client", "metadata", "--help"],
        ["client", "download-model", "--help"],
    ],
)
def test_client_cli_basic(args):
    """
    Test that client specific subcommands exist
    """
    runner = CliRunner()
    out = runner.invoke(cli.gordo, args=args)
    assert (
        out.exit_code == 0
    ), f"Expected output code 0 got '{out.exit_code}', {out.output}"


def test_client_cli_metadata(watchman_service, tmp_dir):
    """
    Test proper execution of client predict sub-command
    """
    runner = CliRunner()

    # Simple metadata fetching
    out = runner.invoke(
        cli.gordo,
        args=[
            "client",
            "--project",
            tu.GORDO_PROJECT,
            "--target",
            tu.GORDO_SINGLE_TARGET,
            "metadata",
        ],
    )
    assert out.exit_code == 0
    assert tu.GORDO_SINGLE_TARGET in out.output

    # Save metadata to file
    output_file = os.path.join(tmp_dir, "metadata.json")
    out = runner.invoke(
        cli.gordo,
        args=[
            "client",
            "--project",
            tu.GORDO_PROJECT,
            "--target",
            tu.GORDO_SINGLE_TARGET,
            "metadata",
            "--output-file",
            output_file,
        ],
    )
    assert out.exit_code == 0, f"{out.exc_info}"
    assert os.path.exists(output_file)
    with open(output_file) as f:
        metadata = json.load(f)
        assert tu.GORDO_SINGLE_TARGET in metadata


def test_client_cli_download_model(watchman_service, tmp_dir):
    """
    Test proper execution of client predict sub-command
    """
    runner = CliRunner()

    # Empty output directory before downloading
    assert len(os.listdir(tmp_dir)) == 0

    out = runner.invoke(
        cli.gordo,
        args=[
            "client",
            "--project",
            tu.GORDO_PROJECT,
            "--target",
            tu.GORDO_SINGLE_TARGET,
            "download-model",
            tmp_dir,
        ],
    )
    assert (
        out.exit_code == 0
    ), f"Expected output code 0 got '{out.exit_code}', {out.output}"

    # Output directory should not be empty any longer
    assert len(os.listdir(tmp_dir)) > 0

    model_output_dir = os.path.join(tmp_dir, tu.GORDO_SINGLE_TARGET)
    assert os.path.isdir(model_output_dir)

    model = serializer.load(model_output_dir)
    assert isinstance(model, BaseEstimator)


@pytest.mark.parametrize(
    "forwarder_args",
    [["--influx-uri", tu.INFLUXDB_URI, "--forward-resampled-sensors"], None],
)
@pytest.mark.parametrize("output_dir", [True, False])
@pytest.mark.parametrize("use_parquet", (True, False))
@pytest.mark.parametrize("session_config", ({}, {"headers": {}}))
def test_client_cli_predict(
    influxdb,
    gordo_name,
    watchman_service,
    tmp_dir,
    forwarder_args,
    trained_model_directory,
    output_dir,
    use_parquet,
    session_config,
):
    """
    Test ability for client to get predictions via CLI
    """
    runner = CliRunner()

    args = ["client", "--metadata", "key,value", "--project", tu.GORDO_PROJECT]
    if session_config:
        args.extend(["--session-config", json.dumps(session_config)])

    args.extend(
        [
            "predict",
            "--parquet" if use_parquet else "--no-parquet",
            "2016-01-01T00:00:00Z",
            "2016-01-01T01:00:00Z",
        ]
    )

    influx_client = client_utils.influx_client_from_uri(
        uri=tu.INFLUXDB_URI, dataframe_client=True
    )
    query = """
        SELECT *
        FROM "resampled"
        """

    # Do we have forwarder args?
    if forwarder_args is not None:
        args.extend(forwarder_args)
        vals = influx_client.query(query)
        # There is no data there before we start doing things
        assert len(vals) == 0

    # Should it write out the predictions to dataframes in an output directory?
    if output_dir:
        args.extend(["--output-dir", tmp_dir])

    # Do we have a data provider, POST else GET requests
    args.extend(
        ["--data-provider", json.dumps(providers.RandomDataProvider().to_dict())]
    )

    # Run without any error
    out = runner.invoke(cli.gordo, args=args)
    assert out.exit_code == 0, f"{out.output}"

    # If we activated forwarder and we had any actual data then there should
    # be resampled values in the influx
    if forwarder_args:
        vals = influx_client.query(query)
        assert len(vals) == 1
        assert len(vals["resampled"]) == 48
        influx_client.drop_measurement("resampled")

    # Did it save dataframes to output dir if specified?
    if output_dir:
        assert os.path.exists(os.path.join(tmp_dir, f"{tu.GORDO_SINGLE_TARGET}.csv.gz"))


@pytest.mark.parametrize(
    "should_fail,start_date,end_date",
    [
        (True, "1888-01-01T00:00:00Z", "1888-02-01T01:00:00Z"),  # Fail on bad dates
        (False, "2016-01-01T00:00:00Z", "2016-01-01T01:00:00Z"),  # pass on good dates
    ],
)
def test_client_cli_predict_non_zero_exit(
    should_fail, start_date, end_date, caplog, influxdb, watchman_service
):
    """
    Test ability for client to get predictions via CLI
    """
    runner = CliRunner()

    # Should fail requesting dates which clearly don't exist.
    args = [
        "client",
        "--metadata",
        "key,value",
        "--project",
        tu.GORDO_PROJECT,
        "predict",
        start_date,
        end_date,
    ]

    data_provider = providers.InfluxDataProvider(
        measurement=tu.INFLUXDB_MEASUREMENT, value_name="Value", uri=tu.INFLUXDB_URI
    )

    args.extend(["--data-provider", json.dumps(data_provider.to_dict())])

    # Run without any error
    with caplog.at_level(logging.CRITICAL):
        out = runner.invoke(cli.gordo, args=args)

    if should_fail:
        assert out.exit_code != 0, f"{out.output or out.exception}"
    else:
        assert out.exit_code == 0, f"{out.output or out.exception}"


@pytest.mark.parametrize(
    "config",
    (
        '{"type": "RandomDataProvider", "max_size": 200}',
        '{"type": "InfluxDataProvider", "measurement": "value"}',
    ),
)
def test_data_provider_click_param(config):
    """
    Test click custom param to load a provider from a string config representation
    """
    expected_provider_type = json.loads(config)["type"]
    provider = custom_types.DataProviderParam()(config)
    assert isinstance(provider, getattr(providers, expected_provider_type))

    # Should also be able to take a file path with the json
    with tempfile.NamedTemporaryFile(mode="w") as config_file:
        json.dump(json.loads(config), config_file)
        config_file.flush()

        provider = custom_types.DataProviderParam()(config_file.name)
        assert isinstance(provider, getattr(providers, expected_provider_type))


def _endpoint_metadata(name: str, healthy: bool) -> EndpointMetadata:
    """
    Helper to build a basic EndpointMetadata with only name and healthy fields set
    """
    return EndpointMetadata(
        data={
            "endpoint-metadata": {"metadata": {"name": name}},
            "healthy": healthy,
            "endpoint": "/gordo/v0/test-project",
        }
    )


@pytest.mark.parametrize("tags", [["C", "A", "B", "D"], tu.SENSORS_STR_LIST])
def test_ml_server_dataframe_to_dict_and_back(tags: typing.List[str]):
    """
    Tests the flow of the server creating a dataframe from the model's data, putting into
    a dict of string to df. lists of values, and the client being able to reconstruct it back
    to the original dataframe (less the second level names)
    """
    # Some synthetic data
    original_input = np.random.random((10, len(tags)))
    model_output = np.random.random((10, len(tags)))

    # Convert this data into a dataframe with multi index columns
    df = model_utils.make_base_dataframe(tags, original_input, model_output)

    # Server then converts this into a dict which maps top level names to lists
    serialized = server_utils.dataframe_to_dict(df)

    # Client reproduces this dataframe
    df_clone = server_utils.dataframe_from_dict(serialized)

    # each subset of column under the top level names should be equal
    top_lvl_names = df.columns.get_level_values(0)
    for top_lvl_name in filter(lambda n: n not in ("start", "end"), top_lvl_names):
        assert np.allclose(df[top_lvl_name].values, df_clone[top_lvl_name].values)


@pytest.mark.parametrize(
    "endpoints,target,ignore_unhealthy,expected",
    [
        # One unhealthy + one healthy + no target + ignoring unhealthy = OK
        (
            [_endpoint_metadata("t1", False), _endpoint_metadata("t2", True)],
            None,
            True,
            [_endpoint_metadata("t2", True)],
        ),
        # One unhealthy + one healthy + no target + NOT ignoring unhealthy = ValueError
        (
            [_endpoint_metadata("t1", False), _endpoint_metadata("t2", True)],
            None,
            False,
            ValueError,
        ),
        # One unhealthy + one healthy + target (healthy) + do not ignore unhealthy = OK
        # because the client doesn't care about the unhealthy endpoint with a set target
        (
            [_endpoint_metadata("t1", False), _endpoint_metadata("t2", True)],
            "t2",
            False,
            [_endpoint_metadata("t2", True)],
        ),
        # All unhealthy = ValueError
        # ...even if we're suppose to ignore unhealthy endpoints, because Noen are healthy
        (
            [_endpoint_metadata("t1", False), _endpoint_metadata("t2", False)],
            None,
            True,
            ValueError,
        ),
        # All unhealthy + ignore unhealthy = ValueError
        # ...because we end up with no endpoints
        (
            [_endpoint_metadata("t1", False), _endpoint_metadata("t2", False)],
            None,
            True,
            ValueError,
        ),
        # All healthy = OK
        (
            [_endpoint_metadata("t1", True), _endpoint_metadata("t2", True)],
            None,
            True,
            [_endpoint_metadata("t1", True), _endpoint_metadata("t2", True)],
        ),
        # All healthy + target = OK (should get back only that target)
        (
            [_endpoint_metadata("t1", True), _endpoint_metadata("t2", True)],
            "t2",
            True,
            [_endpoint_metadata("t2", True)],
        ),
        # Want to filter down to one target, but that target is not healthy, ValueError
        (
            [_endpoint_metadata("t1", True), _endpoint_metadata("t2", False)],
            "t2",
            True,
            ValueError,
        ),
    ],
)
def test_client_endpoint_filtering(
    endpoints: typing.List[EndpointMetadata],
    target: typing.Optional[str],
    ignore_unhealthy: typing.Optional[bool],
    expected: typing.List[EndpointMetadata],
):
    if not isinstance(expected, list):
        with pytest.raises(ValueError):
            Client._filter_endpoints(endpoints, target, ignore_unhealthy)
    else:
        filtered_endpoints = Client._filter_endpoints(
            endpoints, target, ignore_unhealthy
        )
        assert (
            expected == filtered_endpoints
        ), f"Not equal: {expected} \n----\n {filtered_endpoints}"


def test_exponential_sleep_time(caplog, watchman_service):
    endpoint = _endpoint_metadata("t1", False)

    start, end = (
        isoparse("2016-01-01T00:00:00+00:00"),
        isoparse("2016-01-01T12:00:00+00:00"),
    )

    with caplog.at_level(logging.CRITICAL):
        with patch(
            "gordo_components.client.client.sleep", return_value=None
        ) as time_sleep:
            client = Client(project=tu.GORDO_PROJECT)

            client._send_prediction_request(
                X=pd.DataFrame([123]),
                y=None,
                chunk=slice(0, 1),
                endpoint=endpoint,
                start=start,
                end=end,
            )

            expected_calls = [call(8), call(16), call(32), call(64), call(128)]
            time_sleep.assert_has_calls(expected_calls)


def test__handle_response_errors():
    """
    Test expected error raising from gordo_components.client.io._handle_response
    """
    resp = requests.Response()
    resp.status_code = 422
    with pytest.raises(HttpUnprocessableEntity):
        _handle_response(resp)

    resp = requests.Response()
    resp.status_code = 403
    with pytest.raises(BadRequest):
        _handle_response(resp)

    resp = requests.Response()
    resp.status_code = 502
    with pytest.raises(IOError):
        _handle_response(resp)



# -*- coding: utf-8 -*-

import logging

import pytest
import subprocess

from gordo_components.server.server import run_cmd
from gordo_components import serializer


logger = logging.getLogger(__name__)


def test_healthcheck_endpoint(base_route, gordo_ml_server_client):
    """
    Test expected behavior of /<gordo-name>/healthcheck
    """
    # Should also be at the very lowest level as well.
    resp = gordo_ml_server_client.get(f"/healthcheck")
    assert resp.status_code == 200

    resp = gordo_ml_server_client.get(f"{base_route}/healthcheck")
    assert resp.status_code == 200

    data = resp.get_json()
    logger.debug(f"Got resulting JSON response: {data}")
    assert "gordo-server-version" in data


def test_response_header_timing(base_route, gordo_ml_server_client):
    """
    Test that the response contains a `Server-Timing` header
    """
    resp = gordo_ml_server_client.get(f"{base_route}/healthcheck")
    assert resp.status_code == 200
    assert "Server-Timing" in resp.headers
    assert "request_walltime_s" in resp.headers["Server-Timing"]


def test_metadata_endpoint(base_route, gordo_ml_server_client):
    """
    Test the expected behavior of /metadata
    """
    resp = gordo_ml_server_client.get(f"{base_route}/metadata")
    assert resp.status_code == 200

    data = resp.get_json()
    assert "metadata" in data
    assert data["metadata"]["user-defined"]["model-name"] == "test-model"


def test_download_model(base_route, gordo_ml_server_client):
    """
    Test we can download a model, loadable via serializer.loads()
    """
    resp = gordo_ml_server_client.get(f"{base_route}/download-model")

    serialized_model = resp.get_data()
    model = serializer.loads(serialized_model)

    # All models have a fit method
    assert hasattr(model, "fit")

    # Models MUST have either predict or transform
    assert hasattr(model, "predict") or hasattr(model, "transform")


def test_run_cmd(monkeypatch):
    """
    Test that execution error catchings work as expected
    """

    # Call command that raises FileNotFoundError, a subclass of OSError
    cmd = ["gumikorn", "gordo_components.server.server:app"]
    with pytest.raises(OSError):
        run_cmd(cmd)

    # Call command that raises a CalledProcessError
    cmd = ["ping", "--bad-option"]
    with pytest.raises(subprocess.CalledProcessError):
        run_cmd(cmd)



# -*- coding: utf-8 -*-

from typing import Dict, Any

from .base import ConfigElement


class DataProvider(ConfigElement):
    """
    Represents a DataProvider
    """

    def __init__(self, type: str = "DataLakeProvider", **kwargs):
        """
        Parameters
        ----------
        type: str
            Type of DataProvider to initialize
        kwargs: dict
            Other parameters which will be provided to the dataprovider on
            initialization
        """
        self.data_provider_type = type
        self.kwargs = kwargs

    def to_dict(self):
        config = {"type": self.data_provider_type}
        config.update(self.kwargs)  # Other optional kwargs, such as threads
        return config

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "DataProvider":
        return cls(**config)



# -*- coding: utf-8 -*-

import os
import io
import logging
import traceback
import timeit
import typing

import pandas as pd

from flask import Blueprint, current_app, g, send_file, make_response, jsonify, request
from flask_restplus import Resource, fields

from gordo_components import __version__, serializer
from gordo_components.server.rest_api import Api
from gordo_components.server import utils as server_utils
from gordo_components.model import utils as model_utils
from gordo_components.dataset.sensor_tag import SensorTag, normalize_sensor_tags
from gordo_components.server import model_io


logger = logging.getLogger(__name__)

base_blueprint = Blueprint("base_model_view", __name__)

api = Api(
    app=base_blueprint,
    title="Gordo Base Model View API Docs",
    version=__version__,
    description="Documentation for the Gordo ML Server",
    default_label="Gordo Endpoints",
)

# POST type declarations
API_MODEL_INPUT_POST = api.model(
    "Prediction - Multiple Samples", {"X": fields.List(fields.List(fields.Float))}
)
API_MODEL_OUTPUT_POST = api.model(
    "Prediction - Output from POST", {"output": fields.List(fields.List(fields.Float))}
)
_tags = {
    fields.String: fields.Float
}  # tags of single prediction record {'tag-name': tag-value}
_single_prediction_record = {
    "start": fields.DateTime,
    "end": fields.DateTime,
    "tags": fields.Nested(_tags),
    "total_abnormality": fields.Float,
}


class BaseModelView(Resource):
    """
    The base model view.

    Both POST and GET should return the same data response, but can vary
    in how it collects the data.

    POST expects data to be provided, GET uses TimeSeriesDataset

    A typical response might look like this::

         {
        'data': [
            {
           'end': ['2016-01-01T00:10:00+00:00'],
           'model-output': [0.0005317790200933814,
                            -0.0001525811239844188,
                            0.0008310950361192226,
                            0.0015755111817270517],
           'original-input': [0.9135588550070414,
                              0.3472517774179448,
                              0.8994921857179736,
                              0.11982773108991263],
           'start': ['2016-01-01T00:00:00+00:00'],
            },
            ...
        ],

     'tags': [{'asset': None, 'name': 'tag-0'},
              {'asset': None, 'name': 'tag-1'},
              {'asset': None, 'name': 'tag-2'},
              {'asset': None, 'name': 'tag-3'}],
     'time-seconds': '0.1937'}
    """

    methods = ["POST"]

    y: pd.DataFrame = None
    X: pd.DataFrame = None

    @property
    def frequency(self):
        """
        The frequency the model was trained with in the dataset
        """
        return pd.tseries.frequencies.to_offset(g.metadata["dataset"]["resolution"])

    @property
    def tags(self) -> typing.List[SensorTag]:
        return normalize_sensor_tags(g.metadata["dataset"]["tag_list"])

    @property
    def target_tags(self) -> typing.List[SensorTag]:
        if "target_tag_list" in g.metadata["dataset"]:
            return normalize_sensor_tags(g.metadata["dataset"]["target_tag_list"])
        else:
            return []

    @api.response(200, "Success", API_MODEL_OUTPUT_POST)
    @api.expect(API_MODEL_INPUT_POST, validate=False)
    @api.doc(params={"X": "Nested or single list of sample(s) to predict"})
    @server_utils.model_required
    @server_utils.extract_X_y
    def post(self):
        """
        Process a POST request by using provided user data
        """
        return self._process_request()

    def _process_request(self):
        """
        Construct a response which fetches model outputs (transformed) as well
        as original input, transformed model input and, if applicable, inverse
        transformed model output.

        Parameters
        ----------
        context: dict
            Current context mapping of the request. Must be dict where
            items are capable of JSON serialization

        Returns
        -------
        flask.Response
        """
        context = dict()  # type: typing.Dict[str, typing.Any]
        context["status-code"] = 200
        context["tags"] = self.tags
        context["target-tags"] = self.target_tags

        data = None
        context: typing.Dict[typing.Any, typing.Any] = dict()
        X = g.X
        process_request_start_time_s = timeit.default_timer()

        try:
            output = model_io.get_model_output(model=g.model, X=X)
        except ValueError as err:
            tb = traceback.format_exc()
            logger.error(
                f"Failed to predict or transform; error: {err} - \nTraceback: {tb}"
            )
            context["error"] = f"ValueError: {str(err)}"
            return make_response((jsonify(context), 400))

        # Model may only be a transformer, probably an AttributeError, but catch all to avoid logging other
        # exceptions twice if it happens.
        except Exception as exc:
            tb = traceback.format_exc()
            logger.error(
                f"Failed to predict or transform; error: {exc} - \nTraceback: {tb}"
            )
            context["error"] = "Something unexpected happened; check your input data"
            return make_response((jsonify(context), 400))

        else:
            get_model_output_time_s = timeit.default_timer()
            logger.debug(
                f"Calculating model output took "
                f"{get_model_output_time_s-process_request_start_time_s} s"
            )
            data = model_utils.make_base_dataframe(
                tags=self.tags,
                model_input=X.values if isinstance(X, pd.DataFrame) else X,
                model_output=output,
                target_tag_list=self.target_tags,
                index=X.index,
            )
            if request.args.get("format") == "parquet":
                return send_file(
                    io.BytesIO(server_utils.dataframe_into_parquet_bytes(data)),
                    mimetype="application/octet-stream",
                )
            else:
                context["data"] = server_utils.dataframe_to_dict(data)
                return make_response(
                    (jsonify(context), context.pop("status-code", 200))
                )


class MetaDataView(Resource):
    """
    Serve model / server metadata
    """

    @server_utils.metadata_required
    def get(self):
        """
        Get metadata about this endpoint, also serves as /healthcheck endpoint
        """
        model_collection_env_var = current_app.config["MODEL_COLLECTION_DIR_ENV_VAR"]
        return {
            "gordo-server-version": __version__,
            "metadata": g.metadata,
            "env": {model_collection_env_var: os.environ.get(model_collection_env_var)},
        }


class DownloadModel(Resource):
    """
    Download the trained model

    suitable for reloading via ``gordo_components.serializer.loads()``
    """

    @api.doc(
        description="Download model, loadable via gordo_components.serializer.loads"
    )
    @server_utils.model_required
    def get(self):
        """
        Responds with a serialized copy of the current model being served.

        Returns
        -------
        bytes
            Results from ``gordo_components.serializer.dumps()``
        """
        serialized_model = serializer.dumps(g.model)
        buff = io.BytesIO(serialized_model)
        return send_file(buff, attachment_filename="model.tar.gz")


api.add_resource(BaseModelView, "/gordo/v0/<gordo_project>/<gordo_name>/prediction")
api.add_resource(
    MetaDataView,
    "/gordo/v0/<gordo_project>/<gordo_name>/metadata",
    "/gordo/v0/<gordo_project>/<gordo_name>/healthcheck",
)
api.add_resource(DownloadModel, "/gordo/v0/<gordo_project>/<gordo_name>/download-model")



import os
import io
import json
import typing
import re
import time
import logging
from contextlib import contextmanager
from typing import List
from threading import Lock

import docker
import numpy as np
import pandas as pd

import responses
import requests
from influxdb import InfluxDBClient
from flask import Request

from gordo_components.model import models
from gordo_components.watchman import server as watchman_server
from gordo_components.dataset.sensor_tag import SensorTag
from gordo_components.dataset.sensor_tag import to_list_of_strings

logger = logging.getLogger(__name__)

SENSORTAG_LIST = [SensorTag(f"tag-{i}", None) for i in range(4)]
SENSORS_STR_LIST = to_list_of_strings(SENSORTAG_LIST)
INFLUXDB_NAME = "testdb"
INFLUXDB_USER = "root"
INFLUXDB_PASSWORD = "root"
INFLUXDB_MEASUREMENT = "sensors"

INFLUXDB_URI = f"{INFLUXDB_USER}:{INFLUXDB_PASSWORD}@localhost:8086/{INFLUXDB_NAME}"

INFLUXDB_FIXTURE_ARGS = (
    SENSORS_STR_LIST,
    INFLUXDB_NAME,
    INFLUXDB_USER,
    INFLUXDB_PASSWORD,
    SENSORS_STR_LIST,
)

GORDO_HOST = "localhost"
GORDO_PROJECT = "gordo-test"
GORDO_TARGETS = ["machine-1"]
GORDO_SINGLE_TARGET = GORDO_TARGETS[0]
TEST_SERVER_MUTEXT = Lock()


def wait_for_influx(max_wait=120, influx_host="localhost:8086"):
    """Waits up to `max_wait` seconds for influx at `influx_host` to start up.

    Checks by pinging inflix at /ping.

    Parameters
    ----------
    influx_host : str
        Where is influx?
    max_wait : int
        How many seconds to wait in total for influx to start up

    Returns
    -------
    bool
        True if influx started up, False if it did not start during the `max_wait`
        period.

    """
    healtcheck_endpoint = f"http://{influx_host}/ping?verbose=true"
    before_time = time.perf_counter()
    influx_ok = False
    while not influx_ok and time.perf_counter() - before_time < max_wait:
        try:
            code = requests.get(
                healtcheck_endpoint, timeout=1, proxies={"https": "", "http": ""}
            ).status_code
            logger.debug(f"Influx gave code {code}")
            influx_ok = code == 200
        except requests.exceptions.ConnectionError:
            influx_ok = False
        time.sleep(0.5)
    if (time.perf_counter() - before_time) < max_wait:
        logger.info("Found that influx started")
        return True
    else:
        logger.warning("Found that influx never started")
        return False


@contextmanager
def watchman(
    host: str,
    project: str,
    targets: typing.List[str],
    model_location: str,
    namespace: str = "default",
):
    """
    # TODO: This is bananas, make into a proper object with context support?

    Mock a deployed watchman deployment

    Parameters
    ----------
    host: str
        Host watchman should pretend to run on
    project: str
        Project watchman should pretend to care about
    targets:
        Targets watchman should pretend to care about
    model_location: str
        Directory of the model to use in the target(s)
    namespace: str
        Namespace for watchman to make requests in.

    Returns
    -------
    None
    """
    from gordo_components.server import server as gordo_ml_server

    with temp_env_vars(MODEL_COLLECTION_DIR=model_location):
        # Create gordo ml servers
        gordo_server_app = gordo_ml_server.build_app()
        gordo_server_app.testing = True
        gordo_server_app = gordo_server_app.test_client()

        def gordo_ml_server_callback(request):
            """
            Redirect calls to a gordo server to reflect what the local testing app gives
            will call the correct path (assuminng only single level paths) on the
            gordo app.
            """
            if request.method in ("GET", "POST"):

                kwargs = dict()
                if request.body:
                    flask_request = Request.from_values(
                        content_length=len(request.body),
                        input_stream=io.BytesIO(request.body),
                        content_type=request.headers["Content-Type"],
                        method=request.method,
                    )
                    if flask_request.json:
                        kwargs["json"] = flask_request.json
                    else:
                        kwargs["data"] = {
                            k: (io.BytesIO(f.read()), f.filename)
                            for k, f in flask_request.files.items()
                        }

                with TEST_SERVER_MUTEXT:
                    resp = getattr(gordo_server_app, request.method.lower())(
                        request.path_url, **kwargs
                    )
                return (
                    200,
                    resp.headers,
                    json.dumps(resp.json) if resp.json is not None else resp.data,
                )

        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:

            # Gordo ML Server requests
            rsps.add_callback(
                responses.GET,
                re.compile(
                    rf".*ambassador.{namespace}.*\/gordo\/v0\/{project}\/.*.\/.*"
                ),
                callback=gordo_ml_server_callback,
                content_type="application/json",
            )
            rsps.add_callback(
                responses.POST,
                re.compile(
                    rf".*ambassador.{namespace}.*\/gordo\/v0\/{project}\/.*.\/.*"
                ),
                callback=gordo_ml_server_callback,
                content_type="application/json",
            )
            rsps.add_callback(
                responses.GET,
                re.compile(rf".*{host}.*\/gordo\/v0\/{project}\/.*.\/.*"),
                callback=gordo_ml_server_callback,
                content_type="application/json",
            )
            rsps.add_callback(
                responses.POST,
                re.compile(rf".*{host}.*\/gordo\/v0\/{project}\/.*.\/.*"),
                callback=gordo_ml_server_callback,
                content_type="application/json",
            )

            rsps.add_passthru("http+docker://")  # Docker
            rsps.add_passthru("http://localhost:8086")  # Local influx
            rsps.add_passthru("http://localhost:8087")  # Local influx

            # Create a watchman test app
            watchman_app = watchman_server.build_app(
                project_name=project,
                project_version="v123",
                target_names=targets,
                namespace=namespace,
                ambassador_host=host,
                listen_to_kubernetes=False,
            )
            watchman_app.testing = True
            watchman_app = watchman_app.test_client()

            def watchman_callback(_request):
                """
                Redirect calls to a gordo endpoint to reflect what the local testing app gives
                """
                headers = {}
                resp = watchman_app.get("/").json
                return 200, headers, json.dumps(resp)

            # Watchman requests
            rsps.add_callback(
                responses.GET,
                re.compile(rf".*{host}.*\/gordo\/v0\/{project}\/$"),
                callback=watchman_callback,
                content_type="application/json",
            )
            yield


@contextmanager
def temp_env_vars(**kwargs):
    """
    Temporarily set the process environment variables
    """
    _env = os.environ.copy()

    for key in kwargs:
        os.environ[key] = kwargs[key]

    yield

    os.environ.clear()
    os.environ.update(_env)


@contextmanager
def influxdatabase(
    sensors: List[SensorTag], db_name: str, user: str, password: str, measurement: str
):
    """
    Setup a docker based InfluxDB with data points from 2016-01-1 until 2016-01-02 by minute

    Returns
    -------
    InfluxDB
        An interface to the running db instance with .reset() for convenient resetting of the
        db to it's default state, (with original sensors)
    """

    client = docker.from_env()

    logger.info("Starting up influx!")
    influx = None
    try:
        influx = client.containers.run(
            image="influxdb:1.7-alpine",
            environment={
                "INFLUXDB_DB": db_name,
                "INFLUXDB_ADMIN_USER": user,
                "INFLUXDB_ADMIN_PASSWORD": password,
            },
            ports={"8086/tcp": "8086"},
            remove=True,
            detach=True,
        )
        if not wait_for_influx(influx_host="localhost:8086"):
            raise TimeoutError("Influx failed to start")

        logger.info(f"Started influx DB: {influx.name}")

        # Create the interface to the running instance, set default state, and yield it.
        db = InfluxDB(sensors, db_name, user, password, measurement)
        db.reset()
        logger.info("STARTED INFLUX INSTANCE")
        yield db

    finally:
        logger.info("Killing influx container")
        if influx:
            influx.kill()
        logger.info("Killed influx container")


class InfluxDB:
    """
    Simple interface to a running influx.
    """

    def __init__(
        self,
        sensors: List[SensorTag],
        db_name: str,
        user: str,
        password: str,
        measurement: str,
    ):
        self.sensors = sensors
        self.db_name = db_name
        self.user = user
        self.password = password
        self.measurement = measurement

    def reset(self):
        """
        Set the db to contain the default data
        """
        # Seed database with some records
        influx_client = InfluxDBClient(
            "localhost",
            8086,
            self.user,
            self.password,
            self.db_name,
            proxies={"http": "", "https": ""},
        )

        # Drop and re-create the database
        influx_client.drop_database(self.db_name)
        influx_client.create_database(self.db_name)

        dates = pd.date_range(
            start="2016-01-01", periods=2880, freq="min"
        )  # Minute intervals for 2 days

        logger.info("Seeding database")
        for sensor in self.sensors:
            logger.info(f"Loading tag: {sensor.name}")
            points = np.random.random(size=dates.shape[0])
            data = [
                {
                    "measurement": self.measurement,
                    "tags": {"tag": sensor.name},
                    "time": f"{date}",
                    "fields": {"Value": point},
                }
                for point, date in zip(points, dates)
            ]
            influx_client.write_points(data)


def get_model(config):
    type = config.get("type", "")
    Model = getattr(models, type, None)
    if Model is None:
        raise ValueError(
            f'Type of model: "{type}" either not provided or not supported'
        )
    return Model(**config)



# -*- coding: utf-8 -*-

import sys  # noqa
import copy
import requests
import logging
import itertools
import typing
from time import sleep
from threading import Lock
from typing import Dict, Any, List, Callable, Optional
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor


import pandas as pd
from sklearn.base import BaseEstimator
from werkzeug.exceptions import BadRequest

from gordo_components import serializer
from gordo_components.client.io import _handle_response
from gordo_components.client.io import HttpUnprocessableEntity
from gordo_components.client.utils import EndpointMetadata, PredictionResult
from gordo_components.dataset.datasets import TimeSeriesDataset
from gordo_components.data_provider.base import GordoBaseDataProvider
from gordo_components.server import utils as server_utils


logger = logging.getLogger(__name__)


class Client:
    """
    Basic client shipped with Gordo

    Enables some basic communication with a deployed Gordo project
    """

    _mutex = Lock()
    endpoints: List[EndpointMetadata] = []

    def __init__(
        self,
        project: str,
        target: typing.Optional[str] = None,
        host: str = "localhost",
        port: int = 443,
        scheme: str = "https",
        gordo_version: str = "v0",
        metadata: typing.Optional[dict] = None,
        data_provider: typing.Optional[GordoBaseDataProvider] = None,
        prediction_forwarder: typing.Optional[
            Callable[[pd.DataFrame, EndpointMetadata, dict, pd.DataFrame], None]
        ] = None,
        batch_size: int = 100000,
        parallelism: int = 10,
        forward_resampled_sensors: bool = False,
        ignore_unhealthy_targets: bool = False,
        n_retries: int = 5,
        use_parquet: bool = False,
        session: requests.Session = requests.Session(),
    ):
        """

        Parameters
        ----------
        project: str
            Name of the project.
        target: Optional[str]
            Target name if desired to only make predictions against one target.
            Leave as None to run predictions against all targets in Watchman.
        host: str
            Host of where to find watchman and other services.
        port: int
            Port to communicate on.
        scheme: str
            The request scheme to use, ie 'https'.
        gordo_version: str
            The version of major gordo the services are using, ie. 'v0'.
        metadata: Optional[dict]
            Arbitrary mapping of key-value pairs to save to influx with
            prediction runs in 'tags' property
        data_provider: Optional[GordoBaseDataProvider]
            The data provider to use for the dataset. If not set, the client
            will fall back to using the GET /prediction endpoint
        prediction_forwarder: Optional[Callable[[pd.DataFrame, EndpointMetadata, dict, pd.DataFrame], None]]
            callable which will take a dataframe of predictions,
            ``EndpointMetadata``, the metadata, and the dataframe of resampled sensor
            values and forward them somewhere.
        batch_size: int
            How many samples to send to the server, only applicable when data
            provider is supplied.
        parallelism: int
            The maximum number of tasks to run at a given time when
            running predictions
        forward_resampled_sensors: bool
            If true then forward resampled sensor values to the prediction_forwarder
        ignore_unhealthy_targets: bool
            Ignore any targets which are unhealthy. Will raise a ``ValueError``
            if the client encounters any unhealthy endpoints, warnings emitted otherwise
        n_retries: int
            Number of times the client should attempt to retry a failed prediction request. Each time the client
            retires the time it sleeps before retrying is exponentially calculated.
        use_parquet: bool
            Pass the data to the server using the parquet protocol. Default is True
            and recommended as it's more efficient for larger batch sizes. If False JSON
            is used for sending the data back and forth.
        session: requests.Session
            The http session object to use for making requests.
        """

        self.base_url = f"{scheme}://{host}:{port}"
        self.watchman_endpoint = f"{self.base_url}/gordo/{gordo_version}/{project}/"
        self.metadata = metadata if metadata is not None else dict()
        self.prediction_forwarder = prediction_forwarder
        self.data_provider = data_provider
        self.use_parquet = use_parquet
        self.session = session

        # Default, failing back to /prediction on http code 422
        self.prediction_path = "/anomaly/prediction"
        self.batch_size = batch_size
        self.parallelism = parallelism
        self.forward_resampled_sensors = forward_resampled_sensors
        self.n_retries = n_retries
        self.query = f"?format={'parquet' if use_parquet else 'json'}"
        self.target = target
        self.ignore_unhealthy_targets = ignore_unhealthy_targets
        self.endpoints = self.get_endpoints()

    def get_endpoints(self) -> List[EndpointMetadata]:
        # Thread safe single access and updating of endpoints.
        with self._mutex:
            endpoints = self._endpoints_from_watchman(self.watchman_endpoint)
            return self._filter_endpoints(
                endpoints=endpoints,
                target=self.target,
                ignore_unhealthy_targets=self.ignore_unhealthy_targets,
            )

    @staticmethod
    def _filter_endpoints(
        endpoints: typing.List[EndpointMetadata],
        target: typing.Optional[str] = None,
        ignore_unhealthy_targets: typing.Optional[bool] = False,
    ) -> typing.List[EndpointMetadata]:
        """
        Based on the current configuration, filter out endpoints which the client
        should not care about.

        Parameters
        ----------
        endpoints: List[EndpointMetadata]
            List of EndpointMetadata objs
        target: Optional[str] (None)
            Name of the target/machine/endpoint we should filter down to
        ignore_unhealthy_targets: Optional[bool] (False)
            Should the client ignore any unhealthy endpoints?

        Returns
        -------
        List[EndpointMetadata]
            The filtered ``EndpointMetadata``s
        """

        original_endpoints = copy.copy(endpoints)

        # Raise an error if we got some unhealty endpoints and weren't suppose to ignore them.
        # otherwise filter out any unhealthy endpoints
        if (
            not target
            and not ignore_unhealthy_targets
            and not all(e.healthy for e in endpoints)
        ):
            raise ValueError(
                f"Flag 'ignore_unhealthy_targets' is set to False and we encountered some "
                f"unhealthy ones: {[ep for ep in endpoints if ep.healthy is False]}"
            )
        else:
            endpoints = [ep for ep in endpoints if ep.healthy]

        # Filter down to single endpoint if requested
        if target:
            endpoints = [ep for ep in endpoints if ep.name == target]

            # And check for single result and that it's healthy
            if len(endpoints) != 1:
                raise ValueError(
                    f"Found {'multiple' if len(endpoints) else 'no'} endpoints matching "
                    f"target name '{target}' in {original_endpoints}"
                )
            if not endpoints[0].healthy:
                raise ValueError(
                    f"The targeted endpoint: '{endpoints[0]}' is unhealthy"
                )

        # finally, raise an error if all this left us without any endpoints
        if not endpoints:
            raise ValueError(
                f"Found no endpoints out of supplied endpoints: {original_endpoints} after filtering"
            )
        return endpoints

    def _endpoints_from_watchman(self, endpoint: str) -> typing.List[EndpointMetadata]:
        """
        Get a list of endpoints by querying Watchman
        """
        resp = self.session.get(endpoint)
        if not resp.ok:
            raise IOError(f"Failed to get endpoints: {repr(resp.content)}")
        return [EndpointMetadata(data) for data in resp.json()["endpoints"]]

    def download_model(self) -> typing.Dict[str, BaseEstimator]:
        """
        Download the actual model(s) from the ML server /download-model

        Returns
        -------
        Dict[str, BaseEstimator]
            Mapping of target name to the model
        """
        models = dict()
        for endpoint in self.endpoints:
            resp = self.session.get(
                f"{self.base_url + endpoint.endpoint}/download-model"
            )
            if resp.ok:
                models[endpoint.name] = serializer.loads(resp.content)
            else:
                raise IOError(f"Failed to download model: '{repr(resp.content)}'")
        return models

    def get_metadata(self, force_refresh: bool = False) -> typing.Dict[str, dict]:
        """
        Get the metadata for each target

        Parameters
        ----------
        force_refresh : bool
            Even if the previous request was cached, make a new request for metadata.

        Returns
        -------
        Dict[str, dict]
            Mapping of target names to their metadata
        """
        if hasattr(self, "metadata_") and not force_refresh:
            return self.metadata_.copy()

        if force_refresh:
            self.endpoints = self.get_endpoints()  # Forced refresh if
        self.metadata_: Dict[str, dict] = {
            ep.name: ep.raw_metadata() for ep in self.endpoints
        }
        return self.metadata_.copy()

    def predict(
        self,
        start: datetime,
        end: datetime,
        refresh_endpoints: bool = False,
        endpoint_names: Optional[List[str]] = None,
    ) -> typing.Iterable[typing.Tuple[str, pd.DataFrame, typing.List[str]]]:
        """
        Start the prediction process.

        Parameters
        ----------
        start: datetime
        end: datetime
        refresh_endpoints : bool
            Before running predictions, refresh the current endpoints. Default
            ``False`` and will use the endpoints obtained during initialization.
        endpoint_names: Optional[List[str]]
            Optionally only target certain endpoints, referring to them by name.

        Returns
        -------
        List[Tuple[str, pandas.core.DataFrame, List[str]]
            A list of tuples, where:
              0th element is the target name
              1st element is the dataframe of the predictions; complete with a DateTime index.
              2nd element is a list of error messages (if any) for running the predictions
        """
        if refresh_endpoints:
            self.endpoints = self.get_endpoints()

        # Select endpoints if the endpoint_names were provided.
        if endpoint_names:
            endpoints = [ep for ep in self.endpoints if ep.name in endpoint_names]
        else:
            endpoints = self.endpoints

        # For every endpoint, start making predictions for the time range
        with ThreadPoolExecutor(max_workers=self.parallelism) as executor:
            jobs = executor.map(
                lambda ep: self.predict_single_endpoint(ep, start, end), endpoints
            )
            return [(j.name, j.predictions, j.error_messages) for j in jobs]

    def predict_single_endpoint(
        self, endpoint: EndpointMetadata, start: datetime, end: datetime
    ) -> PredictionResult:
        """
        Get predictions based on the /prediction POST endpoint of Gordo ML Servers

        Parameters
        ----------
        endpoint: EndpointMetadata
            Named tuple which has 'endpoint' specifying the full url to the base ml server
        start: datetime
        end: datetime

        Returns
        -------
        dict
            Prediction response from /prediction GET
        """

        # Fetch all of the raw data
        X, y = self._raw_data(endpoint, start, end)

        # Forward sensor data
        if self.prediction_forwarder is not None and self.forward_resampled_sensors:
            self.prediction_forwarder(resampled_sensor_data=X)  # type: ignore

        max_indx = len(X.index) - 1  # Maximum allowable index values

        # Start making batch predictions
        with ThreadPoolExecutor(max_workers=self.parallelism) as executor:
            jobs = executor.map(
                lambda i: self._send_prediction_request(
                    X,
                    y,
                    chunk=slice(i, i + self.batch_size),
                    endpoint=endpoint,
                    start=X.index[i],
                    end=X.index[
                        i + self.batch_size
                        if i + self.batch_size <= max_indx
                        else max_indx
                    ],
                ),
                range(0, X.shape[0], self.batch_size),
            )

            # Accumulate the batched predictions
            prediction_dfs = list()
            error_messages: List[str] = list()
            for prediction_result in jobs:
                if prediction_result.predictions is not None:
                    prediction_dfs.append(prediction_result.predictions)
                error_messages.extend(prediction_result.error_messages)

            predictions = (
                pd.concat(prediction_dfs).sort_index()
                if prediction_dfs
                else pd.DataFrame()
            )
        return PredictionResult(
            name=endpoint.name, predictions=predictions, error_messages=error_messages
        )

    def _send_prediction_request(
        self,
        X: pd.DataFrame,
        y: typing.Optional[pd.DataFrame],
        chunk: slice,
        endpoint: EndpointMetadata,
        start: datetime,
        end: datetime,
    ):
        """
        Post a slice of data to the endpoint

        Parameters
        ----------
        X: pandas.core.DataFrame
            The data for the model, in pandas representation
        chunk: slice
            The slice to take from DataFrame.iloc for the batch size
        endpoint: EndpointMetadata
        start: datetime
        end: datetime

        Notes
        -----
        PredictionResult.predictions may be None if the prediction process fails

        Returns
        -------
        PredictionResult
        """

        kwargs: Dict[str, Any] = dict(
            url=f"{self.base_url + endpoint.endpoint}{self.prediction_path}{self.query}"
        )

        # We're going to serialize the data as either JSON or Arrow
        if self.use_parquet:
            kwargs["files"] = {
                "X": server_utils.dataframe_into_parquet_bytes(X.iloc[chunk]),
                "y": server_utils.dataframe_into_parquet_bytes(y.iloc[chunk])
                if y is not None
                else None,
            }
        else:
            kwargs["json"] = {
                "X": server_utils.dataframe_to_dict(X.iloc[chunk]),
                "y": server_utils.dataframe_to_dict(y.iloc[chunk])
                if y is not None
                else None,
            }

        # Start attempting to get predictions for this batch
        for current_attempt in itertools.count(start=1):
            try:
                try:
                    resp = _handle_response(self.session.post(**kwargs))
                except HttpUnprocessableEntity:
                    self.prediction_path = "/prediction"
                    kwargs[
                        "url"
                    ] = f"{self.base_url + endpoint.endpoint}{self.prediction_path}{self.query}"
                    resp = _handle_response(self.session.post(**kwargs))
            # If it was an IO or TimeoutError, we can retry
            except (
                IOError,
                TimeoutError,
                BadRequest,
                requests.ConnectionError,
                requests.HTTPError,
            ) as exc:
                if current_attempt <= self.n_retries:
                    time_to_sleep = min(2 ** (current_attempt + 2), 300)
                    logger.warning(
                        f"Failed to get response on attempt {current_attempt} out of {self.n_retries} attempts."
                    )
                    sleep(time_to_sleep)
                    continue
                else:
                    msg = (
                        f"Failed to get predictions for dates {start} -> {end} "
                        f"for target: '{endpoint.name}' Error: {exc}"
                    )
                    logger.error(msg)

                    return PredictionResult(
                        name=endpoint.name, predictions=None, error_messages=[msg]
                    )

            # No point in retrying a BadRequest
            except BadRequest as exc:
                msg = (
                    f"Failed with BadRequest error for dates {start} -> {end} "
                    f"for target: '{endpoint.name}' Error: {exc}"
                )
                logger.error(msg)
                return PredictionResult(
                    name=endpoint.name, predictions=None, error_messages=[msg]
                )

            # Process response and return if no exception
            else:

                predictions = self.dataframe_from_response(resp)

                # Forward predictions to any other consumer if registered.
                if self.prediction_forwarder is not None:
                    self.prediction_forwarder(  # type: ignore
                        predictions=predictions,
                        endpoint=endpoint,
                        metadata=self.metadata,
                    )
                return PredictionResult(
                    name=endpoint.name, predictions=predictions, error_messages=[]
                )

    def _raw_data(
        self, endpoint: EndpointMetadata, start: datetime, end: datetime
    ) -> pd.DataFrame:
        """
        Fetch the required raw data in this time range which would
        satisfy this endpoint's /prediction POST

        Parameters
        ----------
        endpoint: EndpointMetadata
            Named tuple representing the endpoint info from Watchman
        start: datetime
        end: datetime

        Returns
        -------
        pandas.core.DataFrame
            Dataframe of required tags and index reflecting the datetime point
        """

        # We want to adjust for any model offset. If the model outputs less than it got in, it requires
        # extra data than what we're being asked to get predictions for.
        # just to give us some buffer zone.
        start = self._adjust_for_offset(
            dt=start,
            resolution=endpoint.resolution,
            n_intervals=endpoint.model_offset + 5,
        )
        dataset = TimeSeriesDataset(
            data_provider=self.data_provider,  # type: ignore
            from_ts=start,
            to_ts=end,
            resolution=endpoint.resolution,
            tag_list=endpoint.tag_list,
            target_tag_list=endpoint.target_tag_list,
        )
        return dataset.get_data()

    @staticmethod
    def _adjust_for_offset(dt: datetime, resolution: str, n_intervals: int = 100):
        """
        Adjust the given date by multiplying ``n_intervals`` by ``resolution``. Such that
        a date of 12:00:00 with ``n_intervals=2`` and ``resolution='10m'`` (10 minutes)
        would result in 11:40

        Parameters
        ----------
        dt: datetime
            Initial datetime to adjust.
        resolution: str
            A string code capable of being parsed by :meth::`pandas.Timedelta`.
        n_intervals: int
            Number of resolution steps to take earlier than the given date.

        Returns
        -------
        datetime
            The new offset datetime object.

        Examples
        --------
        >>> import dateutil
        >>> date = dateutil.parser.isoparse("2019-01-01T12:00:00+00:00")
        >>> offset_date = Client._adjust_for_offset(dt=date, resolution='15m', n_intervals=5)
        >>> str(offset_date)
        '2019-01-01 10:45:00+00:00'
        """
        return dt - (pd.Timedelta(resolution) * n_intervals)

    @staticmethod
    def dataframe_from_response(response: typing.Union[dict, bytes]) -> pd.DataFrame:
        """
        The response from the server, parsed as either JSON / dict or raw bytes,
        of which would be expected to be loadable from :func:`server.utils.dataframe_from_parquet_bytes`

        Parameters
        ----------
        response: Union[dict, bytes]
            The parsed response from the ML server.

        Returns
        -------
        pandas.DataFrame
        """
        if isinstance(response, dict):
            predictions = server_utils.dataframe_from_dict(response["data"])
        else:
            predictions = server_utils.dataframe_from_parquet_bytes(response)
        return predictions


def make_date_ranges(
    start: datetime, end: datetime, max_interval_days: int, freq: str = "H"
):
    """
    Split start and end datetimes into a list of datetime intervals.
    If the interval between start and end is less than ``max_interval_days`` then
    the resulting list will contain the original start & end. ie. [(start, end)]

    Otherwise it will split the intervals by ``freq``, parse-able by pandas.

    Parameters
    ----------
    start: datetime
    end: datetime
    max_interval_days: int
        Maximum days between start and end before splitting into intervals
    freq: str
        String frequency parse-able by Pandas

    Returns
    -------
    List[Tuple[datetime, datetime]]
    """
    if (end - start).days >= max_interval_days:
        # Split into 1hr data ranges
        date_range = pd.date_range(start, end, freq=freq)
        return [
            (date_range[i], date_range[i + 1]) for i in range(0, len(date_range) - 1)
        ]
    else:
        return [(start, end)]



# -*- coding: utf-8 -*-

"""
CLI interfaces
"""

import logging
import sys
import traceback

from gordo_components.dataset.sensor_tag import SensorTagNormalizationError
from gunicorn.glogging import Logger

import jinja2
import yaml
import click
from typing import Dict, Type

from gordo_components.builder.build_model import (
    provide_saved_model,
    check_cache,
    calculate_model_key,
    build_model,
)
from gordo_components.serializer import (
    load_metadata,
    pipeline_into_definition,
    pipeline_from_definition,
)
from gordo_components.server import server
from gordo_components import watchman, __version__
from gordo_components.cli.workflow_generator import workflow_cli
from gordo_components.cli.client import client as gordo_client
from gordo_components.cli.custom_types import key_value_par, HostIP, DataProviderParam

EXCEPTION_TO_EXITCODE: Dict[Type[Exception], int] = {
    PermissionError: 20,
    FileNotFoundError: 30,
    SensorTagNormalizationError: 60,
}

logger = logging.getLogger(__name__)


@click.group("gordo-components")
@click.version_option(version=__version__, message=__version__)
def gordo():
    """
    The main entry point for the CLI interface
    """
    pass


DEFAULT_MODEL_CONFIG = (
    "{'gordo_components.model.models.KerasAutoEncoder': {'kind': "
    "'feedforward_hourglass'}} "
)


@click.command()
@click.argument("name", envvar="MODEL_NAME")
@click.argument("output-dir", default="/data", envvar="OUTPUT_DIR")
@click.argument(
    "model-config", envvar="MODEL_CONFIG", default=DEFAULT_MODEL_CONFIG, type=str
)
@click.argument(
    "data-config",
    envvar="DATA_CONFIG",
    default='{"type": "TimeSeriesDataset"}',
    type=yaml.safe_load,
)
@click.option(
    "--data-provider",
    type=DataProviderParam(),
    envvar="DATA_PROVIDER",
    default=None,
    help="DataProvider dict encoded as json. Must contain a 'type' key with the name of"
    " a DataProvider as value.",
)
@click.option("--metadata", envvar="METADATA", default="{}", type=yaml.safe_load)
@click.option(
    "--model-register-dir",
    default=None,
    envvar="MODEL_REGISTER_DIR",
    type=click.Path(
        exists=False, file_okay=False, dir_okay=True, writable=True, readable=True
    ),
)
@click.option(
    "--print-cv-scores", help="Prints CV scores to stdout", is_flag=True, default=False
)
@click.option(
    "--model-parameter",
    type=key_value_par,
    multiple=True,
    default=(),
    help="Key-Value pair for a model parameter and its value, may use this option "
    "multiple times. Separate key,valye by a comma. ie: --model-parameter key,val "
    "--model-parameter some_key,some_value",
)
@click.option(
    "--evaluation-config",
    envvar="EVALUATION_CONFIG",
    default=yaml.safe_dump(
        {
            "cv_mode": "full_build",
            "scoring_scaler": "sklearn.preprocessing.RobustScaler",
        }
    ),
    type=yaml.safe_load,
)
def build(
    name,
    output_dir,
    model_config,
    data_config,
    data_provider,
    metadata,
    model_register_dir,
    print_cv_scores,
    model_parameter,
    evaluation_config,
):
    """
    Build a model and deposit it into 'output_dir' given the appropriate config
    settings.

    \b
    Parameters
    ----------
    name: str
        Name given to the model to build
    output_dir: str
        Directory to save model & metadata to.
    model_config: str
        String containing a yaml which will be parsed to a dict which will be used in
        initializing the model. Should also contain key 'type' which references the
        model to use. ie. KerasAutoEncoder
    data_config: dict
        kwargs to be used in intializing the dataset. Should also
        contain kwarg 'type' which references the dataset to use. ie. InfluxBackedDataset
    data_provider: str
        A quoted data provider configuration in  JSON/YAML format.
        Should also contain key 'type' which references the data provider to use.

        Example::

          '{"type": "DataLakeProvider", "storename" : "example_store"}'

    metadata: dict
        Any additional metadata to save under the key 'user-defined'
    model_register_dir: path
        Path to a directory which will index existing models and their locations, used
        for re-using old models instead of rebuilding them. If omitted then always
        rebuild
    print_cv_scores: bool
        Print cross validation scores to stdout
    model_parameter: List[Tuple]
        List of model key-values, wheres the values will be injected into the model
        config wherever there is a jinja variable with the key.

    evaluation_config: dict
        Dict of parameters which are exposed to build_model.
            - cv_mode: str
                String which enables three different modes, represented as a key value in evaluation_config:
                * cross_val_only: Only perform cross validation
                * build_only: Skip cross validation and only build the model
                * full_build: Cross validation and full build of the model, default value
                Example::

                    {"cv_mode": "cross_val_only"}
    """
    # Set default data provider for data config
    # TODO: This is for backwards compatibility, as the `data_provider` param should
    # TODO: be provided in the `data_config` itself
    if data_provider is not None:
        data_config["data_provider"] = data_provider

    logger.info(f"Building, output will be at: {output_dir}")
    logger.info(f"Raw model config: {model_config}")
    logger.info(f"Data config: {data_config}")
    logger.info(f"Register dir: {model_register_dir}")

    model_parameter = dict(model_parameter)
    model_config = expand_model(model_config, model_parameter)
    model_config = yaml.full_load(model_config)

    # Convert the config into a pipeline, and back into definition to ensure
    # all default parameters are part of the config.
    logger.debug(f"Ensuring the passed model config is fully expanded.")
    model_config = pipeline_into_definition(pipeline_from_definition(model_config))
    logger.debug(f"Fully expanded model config: {model_config}")

    try:
        if evaluation_config["cv_mode"] == "cross_val_only":

            cache_model_location = None
            if model_register_dir is not None:
                cache_key = calculate_model_key(
                    name,
                    model_config,
                    data_config,
                    evaluation_config,
                    metadata=metadata,
                )
                cache_model_location = check_cache(model_register_dir, cache_key)

            if cache_model_location:
                metadata = load_metadata(cache_model_location)
            else:
                _, metadata = build_model(
                    name, model_config, data_config, metadata, evaluation_config
                )

        else:
            model_location = provide_saved_model(
                name,
                model_config,
                data_config,
                metadata,
                output_dir,
                model_register_dir,
                evaluation_config=evaluation_config,
            )
            metadata = load_metadata(model_location)

        # If the model is cached but without CV scores then we force a rebuild. We do
        # this by deleting the entry in the cache and then rerun
        # `provide_saved_model` (leaving the old model laying around)
        if print_cv_scores:
            retrieved_metadata = metadata
            all_scores = get_all_score_strings(retrieved_metadata)
            if not all_scores:
                logger.warning(
                    "Found that loaded model does not have cross validation values "
                    "even though we were asked to print them, clearing cache and "
                    "rebuilding model"
                )

                model_location = provide_saved_model(
                    name,
                    model_config,
                    data_config,
                    metadata,
                    output_dir,
                    model_register_dir,
                    replace_cache=True,
                    evaluation_config=evaluation_config,
                )
                saved_metadata = load_metadata(model_location)
                all_scores = get_all_score_strings(saved_metadata)

            for score in all_scores:
                print(score)
    except Exception as e:
        exit_code = EXCEPTION_TO_EXITCODE.get(e.__class__, 1)
        traceback.print_exc()
        sys.exit(exit_code)
    else:
        return 0


def expand_model(model_config: str, model_parameters: dict):
    """
    Expands the jinja template which is the model using the variables in
    `model_parameters`

    Parameters
    ----------
    model_config: str
        Jinja template which when expanded becomes a valid model config json.
    model_parameters:
        Parameters for the model config.

    Raises
    ------
    ValueError
        If an undefined variable is used in the model_config.

    Returns
    -------
    str
        The model config with variables expanded

    """
    try:
        model_template = jinja2.Environment(
            loader=jinja2.BaseLoader(), undefined=jinja2.StrictUndefined
        ).from_string(model_config)
        model_config = model_template.render(**model_parameters)
    except jinja2.exceptions.UndefinedError as e:
        raise ValueError("Model parameter missing value!") from e
    logger.info(f"Expanded model config: {model_config}")
    return model_config


def get_all_score_strings(metadata):
    """Given metadata from the model builder this function returns a list of
    strings of the following format:
    {metric_name}-{tag_name}_{fold-fold-number} = {score_val}.  This computes the score for the given tag and
    cross validation split.
    {metric_name}-{tag_name}_{fold-aggregation} = {score_val}. This computes the score for the given tag and aggregates
    the score over all cross validation splits (aggregations currently used are mean, std, min and max)
    {metric_name}_{fold-fold-number} = {score_val}.  This computes the score aggregate across all tags (uses sklearn's default
    aggregation method) for a given cross validation split.
    {metric_name}_{fold-aggregation} = {score_val}.  This computes the score aggregate across all tags (uses sklearn's default
    aggregation method) and cross validation splits (aggregations currently used are mean, std, min and max).

    for katib to pick up.

    Current metric names supported are sklearn score functions: 'r2_score', 'explained_variance_score',
    'mean_squared_error' and 'mean_absolute_error'.  The underscores in such score names are replaced by '-'.

    All spaces in the tag name are also replaced by '-'.

    Parameters
    ----------
    metadata : dict
        Metadata dictionary. Must contain a dictionary in
        metadata.model.cross-validation.scores with at least one metric as key and
        value being another map with score key/values. See example

    Examples
    --------
    >>> score_strings = get_all_score_strings(
    ...  {
    ...     "model": {
    ...         "cross-validation": {
    ...             "scores": {"explained variance tag 0": {"fold_1": 0, "fold_2": 0.9, "fold_3": 0.1,"min": 0.1, "max": 0.9, "mean": 1/3},
    ...                        "explained variance tag 1": {"fold_1": 0.2, "fold_2": 0.3, "fold_3": 0.6, "min": 0.2, "max": 0.6, "mean": 0.3666666666666667},
    ...                        "explained variance" : {"min": 0.1, "max": 0.6, "mean": 0.3499999999999999},
    ...                        "r2 score tag 0" : {"fold_1": 0.8, "fold_2": 0.5, "fold_3": 0.6, "min": 0.5, "max": 0.8, "mean": 0.6333333333333333},
    ...                        "r2 score tag 1" : {"fold_1": 0.4, "fold_2": 0.3, "fold_3": 0.5, "min": 0.3, "max": 0.5, "mean": 0.39999999999999997},
    ...                        "r2 score"  : {"min": 0.4,"max": 0.6, "mean": 0.5166666666666666}
    ...                          }
    ...     }
    ...   }
    ... }
    ... )
    >>> len(score_strings)
    30
    >>> score_strings
    ['explained-variance-tag-0_fold_1=0', 'explained-variance-tag-0_fold_2=0.9', 'explained-variance-tag-0_fold_3=0.1', 'explained-variance-tag-0_min=0.1', 'explained-variance-tag-0_max=0.9', 'explained-variance-tag-0_mean=0.3333333333333333', 'explained-variance-tag-1_fold_1=0.2', 'explained-variance-tag-1_fold_2=0.3', 'explained-variance-tag-1_fold_3=0.6', 'explained-variance-tag-1_min=0.2', 'explained-variance-tag-1_max=0.6', 'explained-variance-tag-1_mean=0.3666666666666667', 'explained-variance_min=0.1', 'explained-variance_max=0.6', 'explained-variance_mean=0.3499999999999999', 'r2-score-tag-0_fold_1=0.8', 'r2-score-tag-0_fold_2=0.5', 'r2-score-tag-0_fold_3=0.6', 'r2-score-tag-0_min=0.5', 'r2-score-tag-0_max=0.8', 'r2-score-tag-0_mean=0.6333333333333333', 'r2-score-tag-1_fold_1=0.4', 'r2-score-tag-1_fold_2=0.3', 'r2-score-tag-1_fold_3=0.5', 'r2-score-tag-1_min=0.3', 'r2-score-tag-1_max=0.5', 'r2-score-tag-1_mean=0.39999999999999997', 'r2-score_min=0.4', 'r2-score_max=0.6', 'r2-score_mean=0.5166666666666666']


    """
    all_scores = []
    for metric_name, scores in (
        metadata.get("model", dict())
        .get("cross-validation", dict())
        .get("scores", dict())
        .items()
    ):
        metric_name = metric_name.replace(" ", "-")
        for score_name, score_val in scores.items():
            score_name = score_name.replace(" ", "-")
            all_scores.append(f"{metric_name}_{score_name}={score_val}")
    return all_scores


@click.command("run-server")
@click.option(
    "--host",
    type=HostIP(),
    help="The host to run the server on.",
    default="0.0.0.0",
    envvar="GORDO_SERVER_HOST",
    show_default=True,
)
@click.option(
    "--port",
    type=click.IntRange(1, 65535),
    help="The port to run the server on.",
    default=5555,
    envvar="GORDO_SERVER_PORT",
    show_default=True,
)
@click.option(
    "--workers",
    type=click.IntRange(1, 4),
    help="The number of worker processes for handling requests.",
    default=2,
    envvar="GORDO_SERVER_WORKERS",
    show_default=True,
)
@click.option(
    "--worker-connections",
    type=click.IntRange(1, 4000),
    help="The maximum number of simultaneous clients per worker process.",
    default=50,
    envvar="GORDO_SERVER_WORKER_CONNECTIONS",
    show_default=True,
)
@click.option(
    "--log-level",
    type=click.Choice(Logger.LOG_LEVELS.keys()),
    help="The log level for the server.",
    default="debug",
    envvar="GORDO_SERVER_LOG_LEVEL",
    show_default=True,
)
def run_server_cli(host, port, workers, worker_connections, log_level):
    """
    Run the gordo server app with Gunicorn
    """
    server.run_server(host, port, workers, worker_connections, log_level.lower())


@click.command("run-watchman")
@click.argument("project-name", envvar="PROJECT_NAME", type=str)
@click.argument("project-version", envvar="PROJECT_VERSION", type=str)
@click.argument("target-names", envvar="TARGET_NAMES", type=yaml.safe_load)
@click.option(
    "--host", type=str, help="The host to run the server on.", default="0.0.0.0"
)
@click.option("--port", type=int, help="The port to run the server on.", default=5555)
@click.option("--debug", type=bool, help="Run in debug mode.", default=False)
@click.option(
    "--namespace",
    type=str,
    help="Namespace watchman should make requests in for ML servers",
    default="kubeflow",
    envvar="NAMESPACE",
)
@click.option(
    "--ambassador-namespace",
    type=str,
    help="Namespace watchman expects Ambassador to be in.",
    default="ambassador",
    envvar="AMBASSADOR_NAMESPACE",
)
@click.option(
    "--ambassador-host",
    type=str,
    help="Full hostname of ambassador. If this is set then `--ambassador-namespace` is "
    "ignored even if set explicitly.",
    default=None,
    envvar="AMBASSADOR_HOST",
)
def run_watchman_cli(
    project_name,
    project_version,
    target_names,
    host,
    port,
    debug,
    namespace,
    ambassador_namespace,
    ambassador_host,
):
    """
    Start the Gordo Watchman server for this project. Which is responsible
    for dynamically comparing expected URLs derived from a project config fle
    against those actually deployed to determine and report their health.

    \b
    Must have the following environment variables set:
        PROJECT_NAME: project_name for the config file
        TARGET_NAMES: A list of non-sanitized machine / target names
    """
    watchman.server.run_server(
        host,
        port,
        debug,
        project_name,
        project_version,
        target_names,
        namespace=namespace,
        ambassador_host=ambassador_host
        if ambassador_host
        else f"ambassador.{ambassador_namespace}",
    )


gordo.add_command(workflow_cli)
gordo.add_command(build)
gordo.add_command(run_server_cli)
gordo.add_command(run_watchman_cli)
gordo.add_command(gordo_client)

if __name__ == "__main__":
    gordo()



# -*- coding: utf-8 -*-

import asyncio
import os
import logging
import tempfile
from typing import List

import pytest
import yaml
import numpy as np

from gordo_components import serializer
from gordo_components.dataset.sensor_tag import SensorTag
from gordo_components.data_provider.providers import InfluxDataProvider
from gordo_components.server import server

from tests import utils as tu


logger = logging.getLogger(__name__)


def pytest_collection_modifyitems(items):
    """
    Update all tests which use influxdb to be marked as a dockertest
    """
    for item in items:
        if hasattr(item, "fixturenames") and "influxdb" in item.fixturenames:
            item.add_marker(pytest.mark.dockertest)


@pytest.fixture(autouse=True)
def check_event_loop():
    loop = asyncio.get_event_loop()
    if loop.is_closed():
        logger.info("Creating new event loop!")
        asyncio.set_event_loop(asyncio.new_event_loop())


@pytest.fixture(scope="session")
def sensors():
    return tu.SENSORTAG_LIST


@pytest.fixture
def tmp_dir():
    """
    Fixture: Temporary directory, removed after test completion
    """
    with tempfile.TemporaryDirectory() as tmp_dir:
        yield tmp_dir


@pytest.fixture(scope="session")
def gordo_name():
    # One gordo-target name
    return tu.GORDO_TARGETS[0]


@pytest.fixture(scope="session")
def gordo_project():
    return tu.GORDO_PROJECT


@pytest.fixture(scope="session")
def api_version():
    return "v0"


@pytest.fixture(scope="session")
def base_route(api_version, gordo_project, gordo_name):
    return f"/gordo/{api_version}/{gordo_project}/{gordo_name}"


@pytest.fixture(scope="session")
def trained_model_directory(
    gordo_project: str, gordo_name: str, sensors: List[SensorTag]
):
    """
    Fixture: Train a basic AutoEncoder and save it to a given directory
    will also save some metadata with the model
    """
    with tempfile.TemporaryDirectory() as model_dir:

        # This is a model collection directory
        collection_dir = os.path.join(model_dir, gordo_project)

        # Model specific to the model being trained here
        model_dir = os.path.join(collection_dir, gordo_name)
        os.makedirs(model_dir, exist_ok=True)

        definition = yaml.safe_load(
            """
            gordo_components.model.anomaly.diff.DiffBasedAnomalyDetector:
                base_estimator:
                    sklearn.pipeline.Pipeline:
                        steps:
                            - sklearn.preprocessing.data.MinMaxScaler
                            - gordo_components.model.models.KerasAutoEncoder:
                                kind: feedforward_hourglass
                        memory:
            """
        )
        model = serializer.pipeline_from_definition(definition)
        X = np.random.random(size=len(sensors) * 10).reshape(10, len(sensors))
        model.fit(X, X)
        serializer.dump(
            model,
            model_dir,
            metadata={
                "dataset": {
                    "tag_list": sensors,
                    "resolution": "10T",
                    "target_tag_list": sensors,
                },
                "name": "machine-1",
                "model": {"model-offset": 0},
                "user-defined": {"model-name": "test-model"},
            },
        )
        yield collection_dir


@pytest.fixture(
    # Data Provider(s) per test requiring this client
    params=[
        InfluxDataProvider(
            measurement="sensors",
            value_name="Value",
            proxies={"https": "", "http": ""},
            uri=tu.INFLUXDB_URI,
        )
    ],
    scope="session",
)
def gordo_ml_server_client(request, trained_model_directory):

    with tu.temp_env_vars(MODEL_COLLECTION_DIR=trained_model_directory):

        app = server.build_app()
        app.testing = True

        yield app.test_client()


@pytest.fixture(scope="session")
def base_influxdb(
    sensors,
    db_name=tu.INFLUXDB_NAME,
    user=tu.INFLUXDB_USER,
    password=tu.INFLUXDB_PASSWORD,
    measurement=tu.INFLUXDB_MEASUREMENT,
):
    """
    Fixture to yield a running influx container and pass a test.utils.InfluxDB
    object which can be used to reset the db to it's original data state.
    """
    with tu.influxdatabase(
        sensors=sensors,
        db_name=db_name,
        user=user,
        password=password,
        measurement=measurement,
    ) as db:
        yield db


@pytest.fixture
def influxdb(base_influxdb):
    """
    Fixture to take a running influx and do a reset after each test to ensure
    the data state is the same for each test.
    """
    logger.info("DOING A RESET ON INFLUX DATA")
    base_influxdb.reset()


@pytest.fixture(scope="module")
def watchman_service(
    trained_model_directory,
    host=tu.GORDO_HOST,
    project=tu.GORDO_PROJECT,
    targets=tu.GORDO_TARGETS,
):
    with tu.watchman(
        host=host,
        project=project,
        targets=targets,
        model_location=trained_model_directory,
    ):
        yield


@pytest.fixture(scope="session")
def repo_dir():
    """
    Return the repository directory for gordo infrastructure
    """
    return os.path.join(os.path.dirname(__file__), "..")



# -*- coding: utf-8 -*-

import ast
import json
import logging
import unittest
from io import StringIO
from datetime import datetime, timezone

import yaml

from gordo_components.workflow.config_elements.dataset import Dataset
from gordo_components.workflow.config_elements.machine import Machine
from gordo_components.workflow.config_elements.normalized_config import NormalizedConfig
from gordo_components.workflow.workflow_generator.workflow_generator import (
    get_dict_from_yaml,
)


logger = logging.getLogger(__name__)


class DatasetConfigElementTestCase(unittest.TestCase):
    def test_from_config(self):
        """
        Test ability to create a Dataset from a config element
        """
        element_str = """
            name: ct-23-0002
            dataset:
              resolution: 2T
              tags:
                - GRA-YE  -23-0751X.PV
                - GRA-TE  -23-0698.PV
                - GRA-PIT -23-0619B.PV
              train_start_date: 2011-05-20T01:00:04+02:00
              train_end_date: 2018-05-10T15:05:50+02:00
        """
        dataset_config = get_dict_from_yaml(StringIO(element_str))["dataset"]
        dataset = Dataset.from_config(dataset_config.copy())
        dataset_config.update({"type": "TimeSeriesDataset"})
        dataset_config["train_start_date"] = dataset_config[
            "train_start_date"
        ].isoformat()
        dataset_config["train_end_date"] = dataset_config["train_end_date"].isoformat()

        # target_tag_list wasn't specified, but should default to the 'tags'
        dataset_config["target_tag_list"] = [
            "GRA-YE  -23-0751X.PV",
            "GRA-TE  -23-0698.PV",
            "GRA-PIT -23-0619B.PV",
        ]
        self.assertEqual(dataset.to_dict(), dataset_config)

    def test_dataset_from_config_checks_dates(self):
        """
        A dataset needs to have train_start_date properly before train_end_date
        """
        element_str = """
            dataset:
              resolution: 2T
              tags:
                - GRA-YE  -23-0751X.PV
                - GRA-TE  -23-0698.PV
                - GRA-PIT -23-0619B.PV
              train_start_date: 2018-05-10T15:05:50+02:00
              train_end_date: 2018-05-10T15:05:50+02:00
        """
        dataset_config = yaml.load(element_str, Loader=yaml.FullLoader)["dataset"]
        with self.assertRaises(ValueError):
            Dataset.from_config(dataset_config)


class MachineConfigElementTestCase(unittest.TestCase):
    default_globals = dict(NormalizedConfig.DEFAULT_CONFIG_GLOBALS)
    # set something different here so we dont have to change the test every time we
    # change the default runtime parameters
    default_globals["runtime"] = {
        "server": {
            "resources": {
                "requests": {"memory": 1, "cpu": 2},
                "limits": {"memory": 3, "cpu": 4},
            }
        }
    }
    default_globals["dataset"] = {"asset": "global-asset"}

    def test_from_config(self):
        """
        Test ability to create a Machine from a config element.
        """

        element_str = """
            name: ct-23-0001-machine
            data_provider:
              threads: 10
            dataset:
              tags: [GRA-TE  -23-0733.PV, GRA-TT  -23-0719.PV, GRA-YE  -23-0751X.PV]
              target_tag_list: [GRA-TE -123-456]
              train_start_date: 2018-01-01T09:00:30Z
              train_end_date: 2018-01-02T09:00:30Z
            model:
              sklearn.pipeline.Pipeline:
                steps:
                  - sklearn.preprocessing.data.MinMaxScaler
                  - gordo_components.model.models.KerasAutoEncoder:
                      kind: feedforward_hourglass
            evaluation:
                scoring_scaler: Null
            metadata:
              id: special-id
        """
        self.maxDiff = None
        element = get_dict_from_yaml(StringIO(element_str))
        machine = Machine.from_config(
            element,
            project_name="test-project-name",
            config_globals=self.default_globals,
        )
        logger.info(f"{machine}")
        self.assertIsInstance(machine, Machine)
        self.assertTrue(len(machine.dataset.tags), 3)

        # The metadata of machine should be json serializable
        json.dumps(machine.to_dict()["metadata"])

        # The metadata of machine should be ast.literal_eval-able when cast as a str
        self.assertEqual(
            ast.literal_eval(str(machine.to_dict()["metadata"])),
            machine.to_dict()["metadata"],
        )
        # dictionary representation of the machine expected:
        self.assertEqual(
            machine.to_dict(),
            {
                "name": "ct-23-0001-machine",
                "data_provider": {"type": "DataLakeProvider", "threads": 10},
                "project_name": "test-project-name",
                "dataset": {
                    "type": "TimeSeriesDataset",
                    "tags": [
                        "GRA-TE  -23-0733.PV",
                        "GRA-TT  -23-0719.PV",
                        "GRA-YE  -23-0751X.PV",
                    ],
                    "target_tag_list": ["GRA-TE -123-456"],
                    "train_start_date": datetime(
                        2018, 1, 1, 9, 0, 30, tzinfo=timezone.utc
                    ).isoformat(),
                    "train_end_date": datetime(
                        2018, 1, 2, 9, 0, 30, tzinfo=timezone.utc
                    ).isoformat(),
                    "asset": "global-asset",
                },
                "evaluation": {
                    "cv_mode": "full_build",
                    "scoring_scaler": None,
                    "metrics": [
                        "explained_variance_score",
                        "r2_score",
                        "mean_squared_error",
                        "mean_absolute_error",
                    ],
                },
                "metadata": {
                    "global-metadata": {},
                    "machine-metadata": {"id": "special-id"},
                },
                "model": {
                    "sklearn.pipeline.Pipeline": {
                        "steps": [
                            "sklearn.preprocessing.data.MinMaxScaler",
                            {
                                "gordo_components.model.models.KerasAutoEncoder": {
                                    "kind": "feedforward_hourglass"
                                }
                            },
                        ]
                    }
                },
                "runtime": {
                    "server": {
                        "resources": {
                            "requests": {"memory": 1, "cpu": 2},
                            "limits": {"memory": 3, "cpu": 4},
                        }
                    }
                },
            },
        )

    def test_invalid_model(self):
        """
        Test invalid model with 'step' instead of 'steps'
        """
        element_str = """
            name: ct-23-0001-machine
            data_provider:
              threads: 10
            dataset:
              tags: [GRA-TE  -23-0733.PV, GRA-TT  -23-0719.PV, GRA-YE  -23-0751X.PV]
              target_tag_list: [GRA-TE -123-456]
              train_start_date: 2018-01-01T09:00:30Z
              train_end_date: 2018-01-02T09:00:30Z
            model:
              sklearn.pipeline.Pipeline:
                step:
                  - sklearn.preprocessing.data.MinMaxScaler
                  - gordo_components.model.models.KerasAutoEncoder:
                      kind: feedforward_hourglass
            evaluation:
                scoring_scaler: Null
            metadata:
              id: special-id
        """
        element = get_dict_from_yaml(StringIO(element_str))
        with self.assertRaises(ValueError):
            Machine.from_config(
                element,
                project_name="test-project-name",
                config_globals=self.default_globals,
            )



# -*- coding: utf-8 -*-

import os
import unittest
import pytest
import logging
import tempfile

from click.testing import CliRunner
from unittest import mock


from gordo_components import cli
from gordo_components.cli.cli import (
    expand_model,
    DEFAULT_MODEL_CONFIG,
    EXCEPTION_TO_EXITCODE,
)
from gordo_components.serializer import serializer
from tests.utils import temp_env_vars

import json

DATA_CONFIG = (
    "{"
    ' "type": "RandomDataset",'
    ' "train_start_date": "2015-01-01T00:00:00+00:00", '
    ' "train_end_date": "2015-06-01T00:00:00+00:00",'
    ' "tags": ["TRC1","TRC2"],'
    ' "target_tag_list":["TRC1","TRC2"]'
    "}"
)

MODEL_CONFIG = {"sklearn.decomposition.pca.PCA": {"svd_solver": "auto"}}
MODEL_CONFIG_WITH_PREDICT = {
    "gordo_components.model.models.KerasAutoEncoder": {"kind": "feedforward_hourglass"}
}

logger = logging.getLogger(__name__)


class CliTestCase(unittest.TestCase):
    """
    Test the expected usability of the CLI interface
    """

    def setUp(self):
        self.runner = CliRunner()

    def test_build_env_args(self):
        """
        Instead of passing OUTPUT_DIR directly to CLI, should be able to
        read environment variables
        """

        logger.info(f"MODEL_CONFIG={json.dumps(MODEL_CONFIG)}")

        with tempfile.TemporaryDirectory() as tmpdir:
            with temp_env_vars(
                MODEL_NAME="model-name",
                OUTPUT_DIR=tmpdir,
                DATA_CONFIG=DATA_CONFIG,
                MODEL_CONFIG=json.dumps(MODEL_CONFIG),
            ):
                result = self.runner.invoke(cli.gordo, ["build"])

            self.assertEqual(result.exit_code, 0, msg=f"Command failed: {result}")
            self.assertGreater(
                len(os.listdir(tmpdir)),
                1,
                msg="Building was supposed to create at least two files (model and "
                "metadata) in OUTPUT_DIR, but it did not!",
            )

    def test_build_use_registry(self):
        """
        Using a registry causes the second build of a model to copy the first to the
        new location.
        """

        with tempfile.TemporaryDirectory() as tmpdir:
            output_dir_1 = os.path.join(tmpdir, "dir1")
            output_dir_2 = os.path.join(tmpdir, "dir2")

            with temp_env_vars(
                MODEL_NAME="model-name",
                OUTPUT_DIR=output_dir_1,
                DATA_CONFIG=DATA_CONFIG,
                MODEL_CONFIG=json.dumps(MODEL_CONFIG),
                MODEL_REGISTER_DIR=tmpdir + "/reg",
            ):
                result1 = self.runner.invoke(cli.gordo, ["build"])

            self.assertEqual(result1.exit_code, 0, msg=f"Command failed: {result1}")
            # OUTPUT_DIR is the only difference
            with temp_env_vars(
                MODEL_NAME="model-name",
                OUTPUT_DIR=output_dir_2,
                DATA_CONFIG=DATA_CONFIG,
                MODEL_CONFIG=json.dumps(MODEL_CONFIG),
                MODEL_REGISTER_DIR=tmpdir + "/reg",
            ):
                result2 = self.runner.invoke(cli.gordo, ["build"])
            self.assertEqual(result2.exit_code, 0, msg=f"Command failed: {result2}")

            first_metadata = serializer.load_metadata(output_dir_1)
            second_metadata = serializer.load_metadata(output_dir_2)

            # The metadata contains the model build date, so if it got rebuilt these two
            # would be different
            assert (
                first_metadata["model"]["model-creation-date"]
                == second_metadata["model"]["model-creation-date"]
            )

    def test_build_use_registry_bust_cache(self):
        """
        Even using a registry we get separate model-paths when we ask for models for
        different configurations.
        """

        with tempfile.TemporaryDirectory() as tmpdir:
            output_dir_1 = os.path.join(tmpdir, "dir1")
            output_dir_2 = os.path.join(tmpdir, "dir2")

            with temp_env_vars(
                MODEL_NAME="model-name",
                OUTPUT_DIR=output_dir_1,
                DATA_CONFIG=DATA_CONFIG,
                MODEL_CONFIG=json.dumps(MODEL_CONFIG),
                MODEL_REGISTER_DIR=tmpdir + "/reg",
            ):
                result1 = self.runner.invoke(cli.gordo, ["build"])

            self.assertEqual(result1.exit_code, 0, msg=f"Command failed: {result1}")

            with temp_env_vars(
                MODEL_NAME="model-name",
                OUTPUT_DIR=output_dir_2,
                # NOTE: Different train dates!
                DATA_CONFIG=(
                    "{"
                    ' "type": "RandomDataset",'
                    ' "train_start_date": "2019-01-01T00:00:00+00:00", '
                    ' "train_end_date": "2019-06-01T00:00:00+00:00",'
                    ' "tags": ["TRC1", "TRC2"],'
                    "}"
                ),
                MODEL_CONFIG=json.dumps(MODEL_CONFIG),
                MODEL_REGISTER_DIR=tmpdir + "/reg",
            ):
                result2 = self.runner.invoke(cli.gordo, ["build"])
            self.assertEqual(result2.exit_code, 0, msg=f"Command failed: {result2}")

            first_metadata = serializer.load_metadata(output_dir_1)
            second_metadata = serializer.load_metadata(output_dir_2)
            # The metadata contains the model build date, so if it got rebuilt these two
            # would be different
            assert (
                first_metadata["model"]["model-creation-date"]
                != second_metadata["model"]["model-creation-date"]
            )

    def test_build_model_with_parameters(self):
        """
        It works to build a simple model with parameters set
        """

        model = """
        {
         "sklearn.decomposition.pca.PCA":
          {
            "svd_solver": "{{svd_solver}}",
            "n_components": {{n_components}}
          }
        }
        """

        svd_solver = "auto"
        n_components = 0.5

        logger.info(f"MODEL_CONFIG={json.dumps(model)}")

        with tempfile.TemporaryDirectory() as tmpdir:
            with temp_env_vars(
                MODEL_NAME="model-name",
                OUTPUT_DIR=tmpdir,
                DATA_CONFIG=DATA_CONFIG,
                MODEL_CONFIG=model,
            ):
                args = [
                    "build",
                    "--model-parameter",
                    f"svd_solver,{svd_solver}",
                    "--model-parameter",
                    f"n_components,{n_components}",
                ]

                # Run it twice to ensure the model location in the location file
                # is only written once and not appended.
                for _ in range(2):

                    result = self.runner.invoke(cli.gordo, args=args)

                    self.assertEqual(
                        result.exit_code, 0, msg=f"Command failed: {result}"
                    )
                    self.assertGreater(
                        len(os.listdir(tmpdir)),
                        1,
                        msg="Building was supposed to create at least two files ("
                        "model and metadata) in OUTPUT_DIR, but it did not!",
                    )

    def test_expand_model_default_works(self):
        self.assertEquals(expand_model(DEFAULT_MODEL_CONFIG, {}), DEFAULT_MODEL_CONFIG)

    def test_expand_model_expand_works(self):
        model_params = {"kind": "hourglass", "num": 5}
        model_template = "{'gordo_components.model.models.KerasAutoEncoder': {'kind': '{{kind}}', 'num': {{num}}}} "
        expected_model = "{'gordo_components.model.models.KerasAutoEncoder': {'kind': 'hourglass', 'num': 5}} "
        self.assertEquals(expand_model(model_template, model_params), expected_model)

    def test_expand_model_complains_on_missing_vars(self):
        model_params = {"kind": "hourglass"}
        model_template = "{'gordo_components.model.models.KerasAutoEncoder': {'kind': '{{kind}}', 'num': {{num}}}} "
        with self.assertRaises(ValueError):
            expand_model(model_template, model_params)


@pytest.mark.parametrize(
    "exception,exit_code",  # ArithmeticError is not in the mapping of exceptions to
    # exit codes, so it should default to 1
    [(FileNotFoundError, 30), (Exception, 1), (ArithmeticError, 1)],
)
def test_build_exit_code(exception, exit_code, tmp_dir: str):
    """
    Test that cli build exists with different exit codes for different errors.
    """

    runner = CliRunner()

    logger.info(f"MODEL_CONFIG={json.dumps(MODEL_CONFIG_WITH_PREDICT)}")
    with mock.patch(
        "gordo_components.cli.cli.build_model",
        mock.MagicMock(side_effect=exception, autospec=True, return_value=None),
    ):
        with temp_env_vars(
            MODEL_NAME="model-name",
            OUTPUT_DIR=tmp_dir,
            DATA_CONFIG=DATA_CONFIG,
            MODEL_CONFIG=json.dumps(MODEL_CONFIG_WITH_PREDICT),
        ):
            result = runner.invoke(
                cli.gordo,
                ["build", '--evaluation-config={"cv_mode": "cross_val_only"}'],
            )
            assert result.exit_code == exit_code


@pytest.mark.parametrize(
    "should_save_model, cv_mode",
    [(True, {"cv_mode": "full_build"}), (False, {"cv_mode": "cross_val_only"})],
)
def test_build_cv_mode(should_save_model: bool, cv_mode: str, tmp_dir: str):
    """
    Testing build with cv_mode set to full and cross_val_only. Checks that cv_scores are
    printed and model are only saved when using the default (full) value.
    """

    runner = CliRunner()

    logger.info(f"MODEL_CONFIG={json.dumps(MODEL_CONFIG_WITH_PREDICT)}")

    with temp_env_vars(
        MODEL_NAME="model-name",
        OUTPUT_DIR=tmp_dir,
        DATA_CONFIG=DATA_CONFIG,
        MODEL_CONFIG=json.dumps(MODEL_CONFIG_WITH_PREDICT),
    ):
        result = runner.invoke(
            cli.gordo, ["build", "--print-cv-scores", f"--evaluation-config={cv_mode}"]
        )
        # Checks that the file is empty or not depending on the mode.
        if should_save_model:
            assert len(os.listdir(tmp_dir)) != 0
        else:
            assert len(os.listdir(tmp_dir)) == 0

        # Checks the output contains 'explained-variance_raw-scores'
        assert "r2-score" in result.output
        assert "mean-squared-error" in result.output
        assert "mean-absolute-error" in result.output
        assert "explained-variance-score" in result.output


@pytest.mark.parametrize(
    "should_be_equal,cv_mode_1, cv_mode_2",
    [
        (True, {"cv_mode": "full_build"}, {"cv_mode": "cross_val_only"}),
        (False, {"cv_mode": "cross_val_only"}, {"cv_mode": "cross_val_only"}),
    ],
)
def test_build_cv_mode_cross_val_cache(
    should_be_equal: bool, cv_mode_1: str, cv_mode_2: str, tmp_dir: str
):
    """
    Checks that cv_scores uses cache if ran after a full build. Loads the same model, and can
    print the cv_scores from them.
    """

    runner = CliRunner()

    logger.info(f"MODEL_CONFIG={json.dumps(MODEL_CONFIG)}")

    model_register_dir = f"{os.path.join(tmp_dir, 'reg')}"

    with temp_env_vars(
        MODEL_NAME="model-name",
        OUTPUT_DIR=tmp_dir,
        DATA_CONFIG=DATA_CONFIG,
        MODEL_CONFIG=json.dumps(MODEL_CONFIG),
        MODEL_REGISTER_DIR=model_register_dir,
    ):

        runner.invoke(cli.gordo, ["build", f"--evaluation-config={cv_mode_1}"])
        runner.invoke(cli.gordo, ["build", f"--evaluation-config={cv_mode_2}"])

        if should_be_equal:
            assert os.path.exists(model_register_dir)
        else:
            assert not os.path.exists(model_register_dir)


def test_build_cv_mode_build_only(tmp_dir: str):
    """
    Testing build with cv_mode set to build_only. Checks that OUTPUT_DIR gets a model
    saved to it. It also checks that the metadata contains cv-duration-sec=None and
    cv-scores={}
    """

    runner = CliRunner()

    logger.info(f"MODEL_CONFIG={json.dumps(MODEL_CONFIG)}")

    with temp_env_vars(
        MODEL_NAME="model-name",
        OUTPUT_DIR=tmp_dir,
        DATA_CONFIG=DATA_CONFIG,
        MODEL_CONFIG=json.dumps(MODEL_CONFIG),
    ):

        metadata_file = f"{os.path.join(tmp_dir, 'metadata.json')}"
        runner.invoke(
            cli.gordo, ["build", '--evaluation-config={"cv_mode": "build_only"}']
        )

        # A model has been saved
        assert len(os.listdir(tmp_dir)) != 0
        with open(metadata_file) as f:
            metadata_json = json.loads(f.read())
            assert metadata_json["model"]["cross-validation"]["cv-duration-sec"] is None
            assert metadata_json["model"]["cross-validation"]["scores"] == {}


@pytest.mark.parametrize(
    "arg,value,exception_expected",
    [
        # Valid values
        ("--host", "0.0.0.0", False),
        ("--host", "127.0.0.0", False),
        ("--port", 5555, False),
        ("--workers", 1, False),
        ("--log-level", "info", False),
        ("--log-level", "debug", False),
        # Invalid values
        ("--host", "0.0.0", True),
        ("--port", 0, True),
        ("--port", 70000, True),
        ("--workers", -1, True),
        ("--log-level", "badlevel", True),
    ],
)
def test_gunicorn_execution_hosts(monkeypatch, arg, value, exception_expected):
    """
    Test the validation of input parameters to the `run_server` function via the gordo cli
    """

    runner = CliRunner()
    with mock.patch(
        "gordo_components.server.server.run_server",
        mock.MagicMock(return_value=None, autospec=True),
    ) as m:
        result = runner.invoke(cli.gordo, ["run-server", arg, value])

        assert (
            (result.exit_code != 0) if exception_expected else (result.exit_code == 0)
        )
        assert m.called_once_with(value)


def test_watchman_to_sql_cli():
    runner = CliRunner()
    args = ["workflow", "watchman-to-sql", "--help"]
    result = runner.invoke(cli.gordo, args)
    assert result.exit_code == 0



import logging

from apscheduler.schedulers.blocking import BlockingScheduler
from mock import MagicMock

from gordo_components.watchman.endpoints_status import EndpointStatuses

logger = logging.getLogger(__name__)


def test_EndpointStatuses(mocker):
    """Tests the main workflow of endpointStatuses.

    We mock away at two places,
    gordo_components.watchman.endpoints_status.watch_for_model_server_service
    which listens for kubernetes events is mocked away, and we construct the events
    manually.

    And the job-scheduler is never actually started, so we dont proceed to do any of
    the jobs, we just check that they are added/removed as desired.

    """

    # We will never call start on the scheduler, so we wont actually do any of the
    # scheduled jobs|
    scheduler = BlockingScheduler()
    project_name = "super_project"
    project_version = "101"
    namespace = "somenamespace"
    host = "localhost"
    target_names = ["target 1", "target 2"]
    mocked_watch = mocker.patch(
        "gordo_components.watchman.endpoints_status.watch_for_model_resource"
    )
    eps = EndpointStatuses(
        scheduler=scheduler,
        project_name=project_name,
        ambassador_host=host,
        model_names=target_names,
        project_version=project_version,
        namespace=namespace,
    )

    assert namespace == mocked_watch.call_args[1]["namespace"]
    assert project_name == mocked_watch.call_args[1]["project_name"]
    assert project_version == mocked_watch.call_args[1]["project_version"]
    event_handler = mocked_watch.call_args[1]["event_handler"]

    # Before receiving any events we only have the targets in `target_names`
    cur_status = eps.statuses()
    assert set([ep["target"] for ep in cur_status]) == set(target_names)

    # And none of them are healthy
    assert all([ep["healthy"] is False for ep in cur_status])

    # Lets start adding some events.

    # Target 1 is online!
    # We make a fake event
    mock_event_obj = dict(
        metadata=dict(labels={"applications.gordo.equinor.com/model-name": "target 1"})
    )

    # And we let the caller know about it
    event_handler({"type": "ADDED", "object": mock_event_obj})

    # The job to update target 1 is added to the joblist
    jobs = scheduler.get_jobs()
    assert len(jobs) == 1
    assert scheduler.get_job("update_model_metadata_target 1") is not None

    # Target 2 is up as well
    mock_event_obj["metadata"]["labels"][
        "applications.gordo.equinor.com/model-name"
    ] = "target 2"
    event_handler({"type": "ADDED", "object": mock_event_obj})

    jobs = scheduler.get_jobs()
    assert len(jobs) == 2
    assert scheduler.get_job("update_model_metadata_target 2") is not None

    # Oida, target 1 seems to be removed!
    mock_event_obj["metadata"]["labels"] = {
        "applications.gordo.equinor.com/model-name": "target 1"
    }

    event_handler({"type": "DELETED", "object": mock_event_obj})

    jobs = scheduler.get_jobs()
    assert len(jobs) == 1
    assert scheduler.get_job("update_model_metadata_target 1") is None
    assert scheduler.get_job("update_model_metadata_target 2") is not None



from gordo_components.workflow.watchman_to_sql.watchman_to_sql import (
    _extract_machines_from_watchman_response,
)


def test_extract_machine_from_watchman_response():
    example_response = {
        "endpoints": [
            {
                "healthy": True,
                "endpoint-metadata": {
                    "metadata": {
                        "name": "test_machine",
                        "dataset": {
                            "train_start_date": "2019-05-08T14:10:38Z",
                            "train_end_date": "2019-06-08T14:10:38Z",
                        },
                    }
                },
            },
            {"healthy": False},
        ]
    }

    all_machines = _extract_machines_from_watchman_response(example_response)

    assert len(all_machines) == 2
    working_machines = [machine for machine in all_machines if machine is not None]
    assert len(working_machines) == 1



# -*- coding: utf-8 -*-

import pytest

import pandas as pd
import numpy as np

from gordo_components.client.forwarders import ForwardPredictionsIntoInflux
from gordo_components.client.utils import EndpointMetadata, influx_client_from_uri
import tests.utils as tu


def test_influx_forwarder(influxdb):
    """
    Test that the forwarder creates correct points from a
    multi-indexed series
    """
    endpoint = EndpointMetadata(
        data={
            "endpoint-metadata": {
                "metadata": {
                    "name": "some-target-name",
                    "dataset": {
                        "tag_list": tu.SENSORTAG_LIST,
                        "target_tag_list": tu.SENSORTAG_LIST,
                        "resolution": "10T",
                    },
                    "model": {"model-offset": 0},
                }
            },
            "endpoint": "/some-endpoint",
            "healthy": True,
        }
    )

    # Feature outs which match length of tags
    # These should then be re-mapped to the sensor tag names
    keys = [("name1", i) for i, _ in enumerate(tu.SENSORTAG_LIST)]

    # Feature outs which don't match the length of the tags
    # These will be kept at 0..N as field names
    keys.extend([("name2", i) for i in range(len(tu.SENSORTAG_LIST) * 2)])

    # Assign all keys unique numbers
    columns = pd.MultiIndex.from_tuples(keys)
    index = pd.date_range("2019-01-01", "2019-01-02", periods=4)
    df = pd.DataFrame(columns=columns, index=index)

    # Generate some unique values for each key, and insert it into that column
    for i, key in enumerate(keys):
        df[key] = range(i, i + 4)

    # Create the forwarder and forward the 'predictions' to influx.
    forwarder = ForwardPredictionsIntoInflux(destination_influx_uri=tu.INFLUXDB_URI)
    forwarder.forward_predictions(predictions=df, endpoint=endpoint)

    # Client to manually verify the points written
    client = influx_client_from_uri(tu.INFLUXDB_URI, dataframe_client=True)

    name1_results = client.query("SELECT * FROM name1")["name1"]

    # Should have the tag names as column names since the shape matched
    assert all(c in name1_results.columns for c in ["machine"] + tu.SENSORS_STR_LIST)
    for i, tag in enumerate(tu.SENSORS_STR_LIST):
        assert np.allclose(df[("name1", i)].values, name1_results[tag].values)

    # Now check the other top level name "name2" is a measurement with the correct points written
    name2_results = client.query("SELECT * FROM name2")["name2"]

    # Should not have the same names as tags, since shape was 2x as long, should just be numeric columns
    assert all(
        [
            str(c) in name2_results.columns
            for c in ["machine"] + list(range(len(tu.SENSORTAG_LIST) * 2))
        ]
    )
    for key in filter(lambda k: k[0] == "name2", keys):
        assert np.allclose(df[key].values, name2_results[str(key[1])].values)



# -*- coding: utf-8 -*-

from typing import Union

import requests
from werkzeug.exceptions import BadRequest


class HttpUnprocessableEntity(BaseException):
    """
    Represents an error from an HTTP status code of 422: UnprocessableEntity.
    Used in our case for calling /anomaly/prediction on a model which does not
    support anomaly behavior.
    """

    pass


def _handle_response(resp: requests.Response) -> Union[dict, bytes]:
    if 200 <= resp.status_code <= 299:
        is_json = resp.headers["content-type"] == "application/json"
        return resp.json() if is_json else resp.content
    else:
        msg = f"Failed to get response: {resp.status_code}: {resp.content!r}"
        if resp.status_code == 422:
            raise HttpUnprocessableEntity()
        elif 400 <= resp.status_code <= 499:
            raise BadRequest(msg)
        else:
            raise IOError(msg)



# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd

from typing import Optional, Union
from datetime import timedelta

from sklearn.preprocessing import RobustScaler
from sklearn.base import BaseEstimator, TransformerMixin

from gordo_components.model.base import GordoBase
from gordo_components.model import utils as model_utils
from gordo_components.model.models import KerasAutoEncoder
from gordo_components.model.anomaly.base import AnomalyDetectorBase


class DiffBasedAnomalyDetector(AnomalyDetectorBase):
    def __init__(
        self,
        base_estimator: BaseEstimator = KerasAutoEncoder(kind="feedforward_hourglass"),
        scaler: TransformerMixin = RobustScaler(),
    ):
        """
        Classifier which wraps a ``base_estimator`` and provides a diff error
        based approach to anomaly detection.

        It trains a ``scaler`` to the target **after** training, purely for
        error calculations. The underlying ``base_estimator`` is trained
        with the original, unscaled, ``y``.

        Parameters
        ----------
        base_estimator: sklearn.base.BaseEstimator
            The model to which normal ``.fit``, ``.predict`` methods will be used.
            defaults to py:class:`gordo_components.model.models.KerasAutoEncoder` with
            ``kind='feedforward_hourglass``
        scaler: sklearn.base.TransformerMixn
            Defaults to ``sklearn.preprocessing.RobustScaler``
            Used for transforming model output and the original ``y`` to calculate
            the difference/error in model output vs expected.
        """
        self.base_estimator = base_estimator
        self.scaler = scaler

    def __getattr__(self, item):
        """
        Treat this model as transparent into base_estimator unless
        referring to something owned by this object
        """
        if item in self.__dict__:
            return getattr(self, item)
        else:
            return getattr(self.base_estimator, item)

    def get_metadata(self):
        if isinstance(self.base_estimator, GordoBase):
            return self.base_estimator.get_metadata()
        else:
            return {
                "scaler": str(self.scaler),
                "base_estimator": str(self.base_estimator),
            }

    def score(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Union[np.ndarray, pd.DataFrame],
        sample_weight: Optional[np.ndarray] = None,
    ) -> float:
        return self.base_estimator.score(X, y)

    def get_params(self, deep=True):
        return {"base_estimator": self.base_estimator, "scaler": self.scaler}

    def fit(self, X: np.ndarray, y: np.ndarray):
        self.base_estimator.fit(X, y)
        self.scaler.fit(y)  # Scaler is used for calculating errors in .anomaly()
        return self

    def anomaly(
        self, X: pd.DataFrame, y: pd.DataFrame, frequency: Optional[timedelta] = None
    ) -> pd.DataFrame:
        """
        Create an anomaly dataframe from the base provided dataframe.

        Parameters
        ----------
        X: pd.DataFrame
            Dataframe representing the data to go into the model.
        y: pd.DataFrame
            Dataframe representing the target output of the model.

        Returns
        -------
        pd.DataFrame
            A superset of the original base dataframe with added anomaly specific
            features
        """

        # Get the model output, falling back to transform if 'predict' doesn't exist
        model_output = (
            self.predict(X) if hasattr(self, "predict") else self.transform(X)
        )

        # Create the basic dataframe with 'model-output' & 'model-input'
        data = model_utils.make_base_dataframe(
            tags=X.columns,
            model_input=getattr(X, "values", X),
            model_output=model_output,
            target_tag_list=y.columns,
            index=getattr(X, "index", None),
            frequency=frequency,
        )

        model_out_scaled = pd.DataFrame(
            self.scaler.transform(data["model-output"]),
            columns=data["model-output"].columns,
            index=data.index,
        )

        # Calculate the total anomaly between all tags for the original/untransformed
        # Ensure to offset the y to match model out, which could be less if it was an LSTM
        scaled_y = self.scaler.transform(y)
        tag_anomaly = np.abs(model_out_scaled - scaled_y[-len(data) :, :])
        tag_anomaly.columns = pd.MultiIndex.from_tuples(
            ("tag-anomaly", col) for col in tag_anomaly.columns
        )

        data = data.join(tag_anomaly)

        # Calculate the total anomaly
        data["total-anomaly"] = np.linalg.norm(data["tag-anomaly"], axis=1)

        return data



# -*- coding: utf-8 -*-

from typing import Dict, Any, List
from datetime import datetime

from .base import ConfigElement
from gordo_components.workflow.config_elements.validators import (
    ValidTagList,
    ValidDatetime,
    ValidDatasetKwargs,
)


class Dataset(ConfigElement):
    """
    Represents a dataset key element by machine within the config file
    and used to create a Dataset from components
    """

    # Required arguments for Dataset
    tags = ValidTagList()
    target_tag_list = ValidTagList()
    train_start_date = ValidDatetime()
    train_end_date = ValidDatetime()

    # Optional kwargs to constructing the dataset, such as 'resolution'
    kwargs = ValidDatasetKwargs()

    def __init__(
        self,
        tags: List[str],
        target_tag_list: List[str],
        train_start_date: datetime,
        train_end_date: datetime,
        **kwargs,
    ):
        """
        Parameters
        ----------
        tags: List[str]
            Tags the dataset consists of.
        train_start_date: datetime
            Earliest possible date (inclusive) of the dataset
        train_end_date: datetime
            Latest possible date of the dataset, must be after `train_start_date`
        kwargs

        """
        self.tags = tags
        self.target_tag_list = target_tag_list
        self.train_start_date = train_start_date
        self.train_end_date = train_end_date
        if train_start_date >= train_end_date:
            raise ValueError(
                f"train_start_date ({train_start_date}) must be before "
                f"train_end_date ({train_end_date})"
            )
        self.kwargs = kwargs

    def to_dict(self):
        config = {
            "type": "TimeSeriesDataset",
            "tags": self.tags,
            "target_tag_list": self.target_tag_list,
            "train_start_date": self.train_start_date.isoformat(),
            "train_end_date": self.train_end_date.isoformat(),
        }
        config.update(self.kwargs)  # Other optional kwargs, such as resolution.

        return config

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "Dataset":

        # Set target_tag_list as tags if it wasn't set; we always expect there to be target tags
        config.setdefault("target_tag_list", config["tags"])

        return cls(**config)



# -*- coding: utf-8 -*-
import logging
from collections import namedtuple
from datetime import datetime
from typing import List, Callable, Optional

import apscheduler.schedulers.base
import kubernetes
import pytz
import requests

from gordo_components.watchman.gordo_k8s_interface import watch_namespaced_custom_object

logger = logging.getLogger(__name__)

EndpointStatus = namedtuple(
    "EndpointStatus",
    ["endpoint", "target", "endpoint_metadata", "healthy", "last_checked", "last_seen"],
)


def endpoint_url_for_model(project_name: str, model_name: str) -> str:
    """Calculates the endpoint URL for a model"""
    return f"gordo/v0/{project_name}/{model_name}/"


def job_name_for_model_update(model_name: str) -> str:
    """ The name of the update model metadata APscheduler job for a model"""
    return f"update_model_metadata_{model_name}"


class EndpointStatuses:
    """
    Represents a interface to getting endpoint metadata / statuses for use
    inside of Watchman. If `listen_to_kubernetes` is true (default and
    recommended) then it will listen to updated from kubernetes, and fetch
    updated metadata when a change is detected. If it is false then we will
    try to retrieve updates every 10 seconds until a success, and then every 5
    minutes after a successful fetch.

    """

    def __init__(
        self,
        scheduler: apscheduler.schedulers.base,
        project_name: str,
        ambassador_host: str,
        model_names: List[str],
        project_version: Optional[str] = None,
        namespace: Optional[str] = None,
        listen_to_kubernetes=True,
    ):
        """

        Parameters
        ----------
        scheduler: apscheduler.schedulers.base
            Scheduler to be used for running misc jobs
        project_name: str
            Project name
        ambassador_host: str
            Full hostname of ambassador
        model_names: List[str]
            List of all model names.
        namespace: Optional[str]
            Namespace to listen for new services in. Required if
            `listen_to_kubernetes`==True
        project_version: Optional[str]
            Project version to listen for updates to. Required if
            `listen_to_kubernetes`==True
        listen_to_kubernetes: bool
            If true then listen for updates from kubernetes (recommended),
            otherwise just try to fetch metadata from the models
            at regular intervals.

        """
        self.model_metadata: dict = {
            model_name: EndpointStatus(
                endpoint=endpoint_url_for_model(project_name, model_name),
                target=model_name,
                healthy=False,
                endpoint_metadata=dict(),
                last_checked=None,
                last_seen=None,
            )
            for model_name in model_names
        }
        self.project_name = project_name
        self.host = ambassador_host
        self.scheduler = scheduler
        if listen_to_kubernetes:
            if namespace is None or project_version is None:
                raise ValueError(
                    "Namespace and project_version required when "
                    "`listen_to_kubernetes`==True"
                )
            # If we are listening to kubernetes events
            # we only do a "manual" update of every endpoint every 20 min
            self.happy_update_model_interval = 1200
            watcher = watch_for_model_resource(
                namespace=namespace,
                project_name=self.project_name,
                project_version=project_version,
                event_handler=self._handle_k8s_model_service_event,
            )

            watcher.start()
        else:
            # If we are not listening to kubernetes events
            # we do a "manual" update of every endpoint every 5 min
            self.happy_update_model_interval = 300
            for model in model_names:
                self._schedule_update_for_model(model, seconds=10)
                self.update_model_metadata(model)

    def statuses(self,) -> List[dict]:
        """
        Return the current List[EndpointStatus] as a list of JSON
        serializable dicts.

        Returns
        -------
        List[dict]
        """

        return [
            {k.replace("_", "-"): v for k, v in es._asdict().items()}
            for es in self.model_metadata.values()
        ]

    def _handle_k8s_model_service_event(self, event: dict):
        """
        Handles a kubernetes model service update event. Either schedules
        an update-job, or removes it if the event is `DELETE`.


        Parameters
        ----------
        event: dict
            Event from kubernetes, see `kubernetes.watch.watch.Watch.stream` for
            description of intended event structure. Needs to have at least a key
            `object`and a key `type`. 'object' should be of type
            `kubernetes.client.models.v1_service.V1Service`

        Returns
        -------
        None
        """
        event_obj: kubernetes.client.models.v1_service.V1Service = event["object"]
        logger.debug(f"Full k8s event: {event}")
        model_name = None
        if event_obj.get("metadata", dict()).get("labels") is not None:
            model_name = event_obj["metadata"]["labels"].get(
                "applications.gordo.equinor.com/model-name", None
            )

        event_type = event["type"]
        logger.info(f"Got K8s event for model {model_name} of type {event_type}")

        if model_name:
            if event_type == "DELETED":
                job_name = job_name_for_model_update(model_name)
                if self.scheduler.get_job(job_name):
                    logger.info(f"Removing job {job_name}")
                    self._unhealty_model(model_name)
                    self.scheduler.remove_job(job_id=job_name)
                else:
                    logger.warning(
                        f"Retrieved a delete event for model {model_name} but had no "
                        f"matching job "
                    )
            else:
                self._schedule_update_for_model(model_name, seconds=2)
        else:
            logger.warning(
                f"Got model-server service notification, but found no model-name. "
                f"Full event: {event} "
            )

    def _unhealty_model(self, model_name):
        """Registers that a model is no longer available"""
        if model_name in self.model_metadata:
            old_model_metadata = self.model_metadata[model_name]
            new_model_metadata = old_model_metadata._replace(
                healthy=False, last_checked=datetime.now(pytz.utc).isoformat()
            )
            self.model_metadata[model_name] = new_model_metadata

    def _schedule_update_for_model(self, model_name, seconds=5):
        """Schedules a regular model-update job for the model, every `seconds`
        seconds. If the job exists it will be updated with the new schedule,
        and if it does not exist it will be added with next_run_time = now
        and interval = `seconds`.

        """
        job_name = job_name_for_model_update(model_name)
        if self.scheduler.get_job(job_name):
            logger.info(
                f"Updating scheduled job {job_name} to run every {seconds} seconds"
            )
            self.scheduler.reschedule_job(
                job_id=job_name, trigger="interval", seconds=seconds
            )
        else:
            logger.info(
                f"Adding scheduled job {job_name} to run every {seconds} seconds"
            )
            self.scheduler.add_job(
                func=self.update_model_metadata,
                trigger="interval",
                seconds=seconds,
                coalesce=True,
                max_instances=1,
                id=job_name,
                next_run_time=datetime.now(),
                kwargs={"model_name": model_name},
            )

    def update_model_metadata(self, model_name: str):
        """
        Updates model metadata for a model, and if it succeeds then reschedule the
        update job to happen every `self.update_model_interval` second. If it fails then
        ensure that we are scheduled to update frequently.

        Parameters
        ----------
        model_name: str
            Name of model to update metadata for.

        """
        logger.info(f"Checking model {model_name}")
        endpoint = fetch_single_model_metadata(
            host=self.host,
            model_name=model_name,
            endpoint_url=endpoint_url_for_model(self.project_name, model_name),
        )
        if endpoint.healthy:
            logger.info(
                f"Found healthy model {model_name}, "
                f"saving metadata and rescheduling update job"
            )
            self.model_metadata[model_name] = endpoint
            self._schedule_update_for_model(
                model_name=model_name, seconds=self.happy_update_model_interval
            )
        else:
            logger.info(f"Found that model {model_name} was not up yet")
            self._unhealty_model(model_name)


def fetch_single_model_metadata(
    host: str, model_name: str, endpoint_url: str
) -> EndpointStatus:
    """
    Fetch metadata for a single model

    Parameters
    ----------
    host: str
        Name of the host to query, e.g. `ambassador`
    model_name: str
        Name of the model
    endpoint_url: str
        Relative (to host) url of the endpoint, e.g. `/gordo/v0/model1`

    Returns
    -------
    EndpointStatus
    """
    endpoint_url = endpoint_url.lstrip("/").rstrip("/")
    base_url = f"http://{host}/{endpoint_url}"
    metadata: dict = dict()
    metadata_resp_ok = False
    try:
        metadata_url = f"{base_url}/metadata"
        logger.info(f"Trying to fetch metadata from url {metadata_url}")
        metadata_resp = requests.get(metadata_url, timeout=2)
        logger.info(f"Url {metadata_url} gave exit code: {metadata_resp.status_code}")
        metadata_resp_ok = metadata_resp.ok
        if metadata_resp_ok:
            metadata = metadata_resp.json()
    except requests.exceptions.RequestException:
        logger.info(f"Failed getting metadata for endpoint {base_url}")
    return EndpointStatus(
        endpoint="/" + endpoint_url,
        target=model_name,
        endpoint_metadata=metadata,
        healthy=metadata_resp_ok,
        last_checked=datetime.now(pytz.utc).isoformat(),
        last_seen=datetime.now(pytz.utc).isoformat() if metadata_resp_ok else None,
    )


def watch_for_model_resource(
    namespace: str, project_name: str, project_version: str, event_handler: Callable
):
    selectors = {
        "applications.gordo.equinor.com/project-name": project_name,
        "applications.gordo.equinor.com/project-version": project_version,
        "app.kubernetes.io/component": "model",
        "app.kubernetes.io/managed-by": "gordo",
        "app.kubernetes.io/name": "model-server",
        "app.kubernetes.io/part-of": "gordo",
    }

    return watch_namespaced_custom_object(
        event_handler=event_handler, namespace=namespace, selectors=selectors
    )



# -*- coding: utf-8 -*-

import io
import pprint
from typing import Tuple, Union, Optional, Dict, List
from collections import namedtuple

from influxdb import DataFrameClient, InfluxDBClient

from gordo_components.dataset.sensor_tag import normalize_sensor_tags, SensorTag


class EndpointMetadata:
    def __init__(self, data: dict):
        """
        Keeps easy access to common endpoint data attributes, the raw data
        being accessible via ``EndpointMetadata.data()``
        """
        self.__data = data

    def raw_metadata(self):
        """
        Access the raw metadata
        """
        return self.__data.copy()

    @property
    def name(self):
        """Name of this endpoint"""
        return self.__data["endpoint-metadata"]["metadata"]["name"]

    @property
    def endpoint(self):
        """
        The path to the endpoint, *not* including the base url
        ie. /gordo/v0/project-name/target-name
        """
        return self.__data["endpoint"]

    @property
    def tag_list(self) -> List[SensorTag]:
        """
        List of the input tags for the model
        """
        return normalize_sensor_tags(
            self.__data["endpoint-metadata"]["metadata"]["dataset"]["tag_list"]
        )

    @property
    def target_tag_list(self) -> List[SensorTag]:
        """
        List of the target tags for the model
        """
        return normalize_sensor_tags(
            self.__data["endpoint-metadata"]["metadata"]["dataset"]["target_tag_list"]
        )

    @property
    def resolution(self):
        """
        Resolution used in aggregation of the data
        """
        return self.__data["endpoint-metadata"]["metadata"]["dataset"]["resolution"]

    @property
    def model_offset(self):
        """
        Any model offset to be expected when getting predictions.
        """
        return self.__data["endpoint-metadata"]["metadata"]["model"].get(
            "model-offset", 0
        )

    @property
    def healthy(self):
        """
        Whether this endpoint is considered available to accept requests.
        """
        return self.__data["healthy"]

    def __eq__(self, other):
        return self.__data == other.__data

    def __repr__(self):
        buff = io.StringIO()
        pprint.pprint(self.__data, stream=buff)
        buff.seek(0)
        return f"EndpointMetadata(data={buff.read()})"


# Prediction result representation, name=str, predictions=dataframe, error_messages=List[str]
PredictionResult = namedtuple("PredictionResult", "name predictions error_messages")


def _parse_influx_uri(uri: str) -> Tuple[str, str, str, str, str, str]:
    """
    Parse an influx URI

    Parameters
    ----------
    uri: str
        Format: <username>:<password>@<host>:<port>/<optional-path>/<db_name>

    Returns
    -------
    (str, str, str, str, str, str)
        username, password, host, port, path, database
    """
    username, password, host, port, *path, db_name = (
        uri.replace("/", ":").replace("@", ":").split(":")
    )
    path_str = "/".join(path) if path else ""
    return username, password, host, port, path_str, db_name


def influx_client_from_uri(
    uri: str,
    api_key: Optional[str] = None,
    api_key_header: Optional[str] = "Ocp-Apim-Subscription-Key",
    recreate: bool = False,
    dataframe_client: bool = False,
    proxies: Dict[str, str] = {"https": "", "http": ""},
) -> Union[InfluxDBClient, DataFrameClient]:
    """
    Get a InfluxDBClient or DataFrameClient from a SqlAlchemy like URI

    Parameters
    ----------
    uri: str
        Connection string format: <username>:<password>@<host>:<port>/<optional-path>/<db_name>
    api_key: str
        Any api key required for the client connection
    api_key_header: str
        The name of the header the api key should be assigned
    recreate: bool
        Re/create the database named in the URI
    dataframe_client: bool
        Return a DataFrameClient instead of a standard InfluxDBClient
    proxies: dict
        A mapping of any proxies to pass to the influx client

    Returns
    -------
    Union[InfluxDBClient, DataFrameClient]
    """

    username, password, host, port, path, db_name = _parse_influx_uri(uri)

    Client = DataFrameClient if dataframe_client else InfluxDBClient

    client = Client(
        host=host,
        port=port,
        database=db_name,
        username=username,
        password=password,
        path=path,
        ssl=bool(api_key),
        proxies=proxies,
    )
    if api_key:
        client._headers[api_key_header] = api_key
    if recreate:
        client.drop_database(db_name)
        client.create_database(db_name)
    return client



# -*- coding: utf-8 -*-

import unittest
import logging
import yaml
import json
import pydoc

from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.decomposition import TruncatedSVD, PCA
from sklearn.preprocessing import MinMaxScaler, FunctionTransformer

from gordo_components.model.models import KerasAutoEncoder
from gordo_components.serializer import (
    pipeline_into_definition,
    pipeline_from_definition,
)
from gordo_components.model.register import register_model_builder

logger = logging.getLogger(__name__)


class PipelineToConfigTestCase(unittest.TestCase):
    def setUp(self):
        self.variations_of_same_pipeline = [
            # Normal
            Pipeline(
                [
                    ("pca1", PCA(n_components=2)),
                    (
                        "fu",
                        FeatureUnion(
                            [
                                ("pca2", PCA(n_components=3)),
                                (
                                    "pipe",
                                    Pipeline(
                                        [
                                            ("minmax", MinMaxScaler()),
                                            ("truncsvd", TruncatedSVD(n_components=2)),
                                        ]
                                    ),
                                ),
                            ]
                        ),
                    ),
                    ("ae", KerasAutoEncoder(kind="feedforward_hourglass")),
                ]
            ),
            # MinMax initialized (wrongly) with a list
            Pipeline(
                [
                    ("pca1", PCA(n_components=2)),
                    (
                        "fu",
                        FeatureUnion(
                            [
                                ("pca2", PCA(n_components=3)),
                                (
                                    "pipe",
                                    Pipeline(
                                        [
                                            ("minmax", MinMaxScaler([0, 1])),
                                            ("truncsvd", TruncatedSVD(n_components=2)),
                                        ]
                                    ),
                                ),
                            ]
                        ),
                    ),
                    ("ae", KerasAutoEncoder(kind="feedforward_hourglass")),
                ]
            ),
            # MinMax initialized with tuple
            Pipeline(
                [
                    ("pca1", PCA(n_components=2)),
                    (
                        "fu",
                        FeatureUnion(
                            [
                                ("pca2", PCA(n_components=3)),
                                (
                                    "pipe",
                                    Pipeline(
                                        [
                                            ("minmax", MinMaxScaler((0, 1))),
                                            ("truncsvd", TruncatedSVD(n_components=2)),
                                        ]
                                    ),
                                ),
                            ]
                        ),
                    ),
                    ("ae", KerasAutoEncoder(kind="feedforward_hourglass")),
                ]
            ),
            # First pipeline without explicit steps param, other with.
            Pipeline(
                [
                    ("pca1", PCA(n_components=2)),
                    (
                        "fu",
                        FeatureUnion(
                            [
                                ("pca2", PCA(n_components=3)),
                                (
                                    "pipe",
                                    Pipeline(
                                        steps=[
                                            ("minmax", MinMaxScaler((0, 1))),
                                            ("truncsvd", TruncatedSVD(n_components=2)),
                                        ]
                                    ),
                                ),
                            ]
                        ),
                    ),
                    ("ae", KerasAutoEncoder(kind="feedforward_hourglass")),
                ]
            ),
        ]

    def test_pipeline_into_definition(self):

        expected_definition = """
            sklearn.pipeline.Pipeline:
                memory: null
                steps:
                    - sklearn.decomposition.pca.PCA:
                        copy: true
                        iterated_power: auto
                        n_components: 2
                        random_state: null
                        svd_solver: auto
                        tol: 0.0
                        whiten: false
                    - sklearn.pipeline.FeatureUnion:
                        n_jobs: null
                        transformer_list:
                        - sklearn.decomposition.pca.PCA:
                            copy: true
                            iterated_power: auto
                            n_components: 3
                            random_state: null
                            svd_solver: auto
                            tol: 0.0
                            whiten: false
                        - sklearn.pipeline.Pipeline:
                            memory: null
                            steps:
                            - sklearn.preprocessing.data.MinMaxScaler:
                                copy: true
                                feature_range:
                                  - 0
                                  - 1
                            - sklearn.decomposition.truncated_svd.TruncatedSVD:
                                algorithm: randomized
                                n_components: 2
                                n_iter: 5
                                random_state: null
                                tol: 0.0
                            verbose: false
                        transformer_weights: null
                        verbose: false
                    - gordo_components.model.models.KerasAutoEncoder:
                        kind: feedforward_hourglass
                verbose: false
            """

        expected_definition = yaml.safe_load(expected_definition)

        for pipe in self.variations_of_same_pipeline:

            definition = pipeline_into_definition(
                pipeline_from_definition(pipeline_into_definition(pipe))
            )

            self.assertEqual(
                json.dumps(definition),
                json.dumps(expected_definition),
                msg=f"Failed output:\n{definition}\nExpected:----------------\n{expected_definition}",
            )

    def test_into_from(self):
        """
        Pass Pipeline into definition, and then from that definition
        """
        from gordo_components.model.transformer_funcs.general import multiply_by

        self.factories = register_model_builder.factories
        for model in self.factories.keys():

            for model_kind in self.factories[model].keys():
                pipe = Pipeline(
                    [
                        ("step_0", PCA(n_components=2)),
                        (
                            "step_1",
                            FeatureUnion(
                                [
                                    ("step_0", PCA(n_components=3)),
                                    (
                                        "step_1",
                                        Pipeline(
                                            steps=[
                                                ("step_0", MinMaxScaler((0, 1))),
                                                (
                                                    "step_1",
                                                    TruncatedSVD(n_components=2),
                                                ),
                                            ]
                                        ),
                                    ),
                                ]
                            ),
                        ),
                        (
                            "step_2",
                            FunctionTransformer(
                                func=multiply_by, kw_args={"factor": 1}
                            ),
                        ),
                        (
                            "step_3",
                            pydoc.locate(f"gordo_components.model.models.{model}")(
                                kind=model_kind
                            ),
                        ),
                    ]
                )

                pipeline_from_definition(pipeline_into_definition(pipe))

    def test_from_into(self):
        """
        Create pipeline from definition, and create from that definition
        """
        self.factories = register_model_builder.factories
        for model in self.factories.keys():
            for model_kind in self.factories[model].keys():
                definition = f"""
                    sklearn.pipeline.Pipeline:
                        steps:
                            - sklearn.decomposition.pca.PCA:
                                n_components: 2
                                copy: true
                                whiten: false
                                svd_solver: auto
                                tol: 0.0
                                iterated_power: auto
                                random_state:
                            - sklearn.preprocessing._function_transformer.FunctionTransformer:
                                func: gordo_components.model.transformer_funcs.general.multiply_by
                                kw_args:
                                    factor: 1
                                inverse_func: gordo_components.model.transformer_funcs.general.multiply_by
                                inv_kw_args:
                                    factor: 1
                            - sklearn.pipeline.FeatureUnion:
                                transformer_list:
                                - sklearn.decomposition.pca.PCA:
                                    n_components: 3
                                    copy: true
                                    whiten: false
                                    svd_solver: auto
                                    tol: 0.0
                                    iterated_power: auto
                                    random_state:
                                - sklearn.pipeline.Pipeline:
                                    steps:
                                    - sklearn.preprocessing.data.MinMaxScaler:
                                        feature_range:
                                        - 0
                                        - 1
                                        copy: true
                                    - sklearn.decomposition.truncated_svd.TruncatedSVD:
                                        n_components: 2
                                        algorithm: randomized
                                        n_iter: 5
                                        random_state:
                                        tol: 0.0
                                    memory:
                                    verbose: false
                                n_jobs: 1
                                transformer_weights:
                                verbose: false
                            - gordo_components.model.models.{model}:
                                kind: {model_kind}
                        memory:
                        verbose: false
                    """
                definition = yaml.safe_load(definition)
                pipe = pipeline_from_definition(definition)
                pipeline_into_definition(pipe)


def test_captures_kwarg_to_init():
    """
    Our models allow kwargs which are put into the underlying keras model or to construct
    the underlying model.
    We want to ensure into defintion captures kwargs which are part of the model
    parameters but not part of the __init__ signature
    """
    ae = KerasAutoEncoder(kind="feedforward_hourglass", some_fancy_param="Howdy!")
    definition = pipeline_into_definition(ae)
    parameters = definition[
        f"{KerasAutoEncoder.__module__}.{KerasAutoEncoder.__name__}"
    ]
    assert "some_fancy_param" in parameters
    assert parameters["some_fancy_param"] == "Howdy!"

    # And make sure we can init again
    KerasAutoEncoder(**parameters)
